import unittest

from orangecontrib.bioinformatics.ncbi.gene import OrangeTableAnnotations
from orangecontrib.bioinformatics.widgets.OWGEODatasets import OWGEODatasets
from Orange.widgets.tests.base import WidgetTest
from Orange.data import Table


class TestOWGEODatasets(WidgetTest):

    def setUp(self):
        self.test_sample = 'GDS1001'
        self.test_organism = '10090'
        self.widget = self.create_widget(OWGEODatasets)

    def test_output_data(self):
        self.widget.auto_commit = True

        gds_info = self.widget.gds_info

        # check if GDSInfo loaded successfully
        self.assertIsNotNone(gds_info.get)

        # set selected gds
        self.widget.selected_gds = gds_info.get(self.test_sample)
        self.widget._set_selection()
        self.widget.on_gds_selection_changed()

        output_data = self.get_output(self.widget.Outputs.gds_data, wait=20000)

        self.assertIsInstance(output_data, Table)
        # test if table on the output is properly annotated
        table_annotations = output_data.attributes

        self.assertTrue(len(table_annotations) > 0)
        self.assertTrue(OrangeTableAnnotations.tax_id in table_annotations)
        self.assertTrue(OrangeTableAnnotations.gene_as_attribute_name in table_annotations)

        # by default genes are in columns not in rows
        self.assertTrue(OrangeTableAnnotations.gene_id_attribute in table_annotations)
        # check if taxonomy is correct
        self.assertTrue(table_annotations[OrangeTableAnnotations.tax_id] == self.test_organism)


if __name__ == '__main__':
    unittest.main()



""" GEO update """
import ftplib
import pickle
import re
import time

from server_update import *
from orangecontrib.bioinformatics.ncbi import taxonomy
from orangecontrib.bioinformatics import geo
from server_update.tests.test_GEO import GEOTest

force_update = False
DOMAIN = 'GEO'
GDS_INFO = 'gds_info.pickled'
TITLE = 'Gene Expression Omnibus data sets information'
TAGS = ['Gene Expression Omnibus', 'data sets', 'GEO', 'GDS']

FTP_NCBI = 'ftp.ncbi.nih.gov'
NCBI_DIR = 'pub/geo/DATA/SOFT/GDS'
domain_path = sf_local.localpath(DOMAIN)
localfile = sf_local.localpath(DOMAIN, GDS_INFO)
create_folder(domain_path)

try:
    sf_local.localpath_download(DOMAIN, GDS_INFO)
    # read the information from the local file
    with open(localfile, 'rb') as f:
        gds_info, excluded = pickle.load(f, encoding='latin1')
        f.close()

except FileNotFoundError as e:
    print('{} file on the server not found!'.format(GDS_INFO))
    force_update = True


# if needed to refresh the data base
if force_update:
    gds_info, excluded = ({}, {})

# list of common organisms may have changed, rescan excluded list
excluded = dict([(id, taxid) for id, taxid in excluded.items()
                 if taxid not in taxonomy.common_taxids()])
excluded.update([(id, info["taxid"]) for id, info in gds_info.items()
                 if info["taxid"] not in taxonomy.common_taxids()])
gds_info = dict([(id, info) for id, info in gds_info.items()
                 if info["taxid"] in taxonomy.common_taxids()])

# get the list of GDS files from NCBI directory
print("Retrieving ftp directory ...")
ftp = ftplib.FTP(FTP_NCBI)
ftp.login()
ftp.cwd(NCBI_DIR)
dirlist = []
ftp.dir(dirlist.append)

m = re.compile("GDS[0-9]*")
gds_names = [m.search(d).group(0) for d in dirlist if m.search(d)]
gds_names = [name for name in gds_names if not (name in gds_info or name in excluded)]
print('{} new files will be added!'.format(len(gds_names)))
skipped = []

helper = SyncHelper(DOMAIN, GEOTest)

if len(gds_names):
    for count, gds_name in enumerate(gds_names):
        print("%3d of %3d -- Adding %s ..." % (count + 1, len(gds_names), gds_name))
        try:
            time.sleep(1)
            gds = geo.GDS(gds_name)
            if gds.info["taxid"] not in taxonomy.common_taxids():
                excluded[gds_name] = gds.info["taxid"]
                print("... excluded (%s)." % gds.info["sample_organism"])
            else:
                gds_info.update({gds_name: gds.info})
                with open(localfile, 'wb') as f:
                    pickle.dump((gds_info, excluded), f, True)
                    f.close()
                print("... added.")
        except Exception as ex:
            print("... skipped (error):", str(ex))
            skipped.append(gds_name)

    # update .info file
    create_info_file(localfile,
                     domain=DOMAIN,
                     filename=GDS_INFO,
                     source=SOURCE_SERVER,
                     title=TITLE,
                     tags=TAGS)

    print("GDS data sets: %d" % len(gds_info))
    print("Organisms:")

    organisms = [info["sample_organism"] for info in gds_info.values()]
    for org in set(organisms):
        print("  %s (%d)" % (org, organisms.count(org)))

    # sync files with remote server
    helper.run_tests()
    helper.sync_files()
    helper.remove_update_folder()

else:
    helper.remove_update_folder()
    sys.exit(up_to_date)




""" GEO module tests """
import unittest

from Orange.data import Table

from orangecontrib.bioinformatics.geo.dataset import GDSInfo, GDS
from orangecontrib.bioinformatics.ncbi.gene import OrangeTableAnnotations


class TestGEO(unittest.TestCase):
    test_sample = 'GDS1001'
    test_organism = '10090'

    def test_gds_info(self):
        gds_info = GDSInfo()
        self.assertIsNotNone(gds_info)
        self.assertGreater(len(gds_info.keys()), 0)
        self.assertGreater(len(gds_info.items()), 0)
        self.assertGreater(len(gds_info.values()), 0)

        self.assertIsNotNone(gds_info[self.test_sample])
        self.assertEqual(gds_info[self.test_sample]['genes'], 9561)
        self.assertEqual(int(gds_info[self.test_sample]['sample_count']), 4)
        self.assertEqual(len(gds_info[self.test_sample]['subsets']), 2)

    def test_gds_data(self):
        gds_info = GDSInfo()
        gds_table = GDS(self.test_sample)

        # test if data is downloaded
        self.assertIsNotNone(gds_table)
        self.assertIsInstance(gds_table, Table)

        # test data table values
        rows, columns = gds_table.X.shape
        self.assertEqual(int(gds_info[self.test_sample]['sample_count']), rows)
        self.assertEqual(int(gds_info[self.test_sample]['genes']), columns)

        # test data table annotations
        self.assertTrue(OrangeTableAnnotations.gene_as_attribute_name in gds_table.attributes)
        self.assertTrue(OrangeTableAnnotations.gene_id_attribute in gds_table.attributes)
        self.assertTrue(OrangeTableAnnotations.tax_id in gds_table.attributes)

        self.assertTrue(gds_table.attributes[OrangeTableAnnotations.gene_as_attribute_name])

    def test_gds_data_transposed(self):
        gds_info = GDSInfo()
        gds_table = GDS(self.test_sample, transpose=True)

        # test if data is downloaded
        self.assertIsNotNone(gds_table)
        self.assertIsInstance(gds_table, Table)

        # test data table values
        rows, columns = gds_table.X.shape
        self.assertEqual(int(gds_info[self.test_sample]['sample_count']), columns)
        self.assertEqual(int(gds_info[self.test_sample]['genes']), rows)

        # test data table annotations
        self.assertTrue(OrangeTableAnnotations.gene_as_attribute_name in gds_table.attributes)
        self.assertTrue(OrangeTableAnnotations.gene_id_column in gds_table.attributes)
        self.assertTrue(OrangeTableAnnotations.tax_id in gds_table.attributes)

        self.assertFalse(gds_table.attributes[OrangeTableAnnotations.gene_as_attribute_name])


if __name__ == '__main__':
    unittest.main()



""" GeneSets """
import threading
import concurrent.futures

from functools import partial
from typing import Optional

from AnyQt.QtWidgets import (
    QTreeView, QHeaderView, QHBoxLayout,
)
from AnyQt.QtCore import (
    Qt, QSize, QThreadPool, Slot, QThread, QItemSelection, QItemSelectionRange, QItemSelectionModel
)
from AnyQt.QtGui import (
    QColor, QStandardItemModel, QStandardItem,
)

from Orange.widgets.gui import (
    vBox, lineEdit, ProgressBar, LinkRole, LinkStyledItemDelegate,
    auto_commit, widgetLabel, spin, comboBox, widgetBox
)
from Orange.data import Domain, Table, DiscreteVariable, StringVariable, filter as table_filter
from Orange.widgets.widget import OWWidget, Msg
from Orange.widgets.settings import Setting
from Orange.widgets.utils.signals import Output, Input
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.utils.concurrent import ThreadExecutor, FutureWatcher, methodinvoke

from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID, GENE_AS_ATTRIBUTE_NAME, GENE_ID_COLUMN, GENE_ID_ATTRIBUTE,
    ERROR_ON_MISSING_ANNOTATION, ERROR_ON_MISSING_GENE_ID, ERROR_ON_MISSING_TAX_ID
)

from orangecontrib.bioinformatics.widgets.utils.gui import GeneSetsSelection
from orangecontrib.bioinformatics.widgets.utils.gui import NumericalColumnDelegate, FilterProxyModel
from orangecontrib.bioinformatics import geneset


class Task:
    future = None
    watcher = None
    cancelled = False

    def cancel(self):
        self.cancelled = True
        self.future.cancel()
        concurrent.futures.wait([self.future])


class OWGeneSets(OWWidget):
    name = "Gene Sets"
    description = ""
    icon = "icons/OWGeneSets.svg"
    priority = 9
    want_main_area = True

    COUNT, GENES, CATEGORY, TERM = range(4)
    DATA_HEADER_LABELS = ["Count", 'Genes In Set', 'Category', 'Term']

    organism = Setting(None, schema_only=True)
    stored_gene_sets_selection = Setting([], schema_only=True)
    selected_rows = Setting([], schema_only=True)
    custom_gene_set_indicator = Setting(None, schema_only=True)

    min_count = Setting(5)
    use_min_count = Setting(True)
    auto_commit = Setting(True)

    class Inputs:
        genes = Input("Data", Table)
        custom_sets = Input('Custom Gene Sets', Table)

    class Outputs:
        matched_genes = Output("Matched Genes", Table)

    class Information(OWWidget.Information):
        pass

    class Warning(OWWidget.Warning):
        all_sets_filtered = Msg('All sets were filtered out.')

    class Error(OWWidget.Error):
        organism_mismatch = Msg('Organism in input data and custom gene sets does not match')
        missing_annotation = Msg(ERROR_ON_MISSING_ANNOTATION)
        missing_gene_id = Msg(ERROR_ON_MISSING_GENE_ID)
        missing_tax_id = Msg(ERROR_ON_MISSING_TAX_ID)
        cant_reach_host = Msg("Host orange.biolab.si is unreachable.")
        cant_load_organisms = Msg("No available organisms, please check your connection.")

    def __init__(self):
        super().__init__()

        # commit
        self.commit_button = None

        # progress bar
        self.progress_bar = None
        self.progress_bar_iterations = None

        # data
        self.input_data = None
        self.input_genes = []
        self.tax_id = None
        self.use_attr_names = None
        self.gene_id_attribute = None
        self.gene_id_column = None

        # custom gene sets
        self.custom_data = None
        self.feature_model = DomainModel(valid_types=(DiscreteVariable, StringVariable))
        self.custom_gs_col_box = None
        self.gs_label_combobox = None
        self.custom_tax_id = None
        self.custom_use_attr_names = None
        self.custom_gene_id_attribute = None
        self.custom_gene_id_column = None
        self.num_of_custom_sets = None

        # Gene Sets widget
        self.gs_widget = None

        # info box
        self.input_info = None
        self.num_of_sel_genes = 0

        # filter
        self.line_edit_filter = None
        self.search_pattern = ''
        self.organism_select_combobox = None

        # data model view
        self.data_view = None
        self.data_model = None

        # gene matcher NCBI
        self.gene_matcher = None

        # filter proxy model
        self.filter_proxy_model = None

        # hierarchy widget
        self.hierarchy_widget = None
        self.hierarchy_state = None

        # spinbox
        self.spin_widget = None

        # threads
        self.threadpool = QThreadPool(self)
        self.workers = None

        self._task = None  # type: Optional[Task]
        self._executor = ThreadExecutor()

        # gui
        self.setup_gui()

    def __reset_widget_state(self):
        self.update_info_box()
        # clear data view
        self.init_item_model()
        # reset filters
        self.setup_filter_model()

    def cancel(self):
        """
        Cancel the current task (if any).
        """
        if self._task is not None:
            self._task.cancel()
            assert self._task.future.done()
            # disconnect the `_task_finished` slot
            self._task.watcher.done.disconnect(self._init_gene_sets_finished)
            self._task = None

    @Slot()
    def progress_advance(self):
        # GUI should be updated in main thread. That's why we are calling advance method here
        if self.progress_bar:
            self.progress_bar.advance()

    def __get_input_genes(self):
        self.input_genes = []

        if self.use_attr_names:
            for variable in self.input_data.domain.attributes:
                self.input_genes.append(str(variable.attributes.get(self.gene_id_attribute, '?')))
        else:
            genes, _ = self.input_data.get_column_view(self.gene_id_column)
            self.input_genes = [str(g) for g in genes]

    def handle_custom_gene_sets(self, select_customs_flag=False):
        if self.custom_gene_set_indicator:
            if self.custom_data is not None and self.custom_gene_id_column is not None:

                if self.__check_organism_mismatch():
                    # self.gs_label_combobox.setDisabled(True)
                    self.Error.organism_mismatch()
                    self.gs_widget.update_gs_hierarchy()
                    return

                if isinstance(self.custom_gene_set_indicator, DiscreteVariable):
                    labels = self.custom_gene_set_indicator.values
                    gene_sets_names = [labels[int(idx)] for idx
                                       in self.custom_data.get_column_view(self.custom_gene_set_indicator)[0]]
                else:
                    gene_sets_names, _ = self.custom_data.get_column_view(self.custom_gene_set_indicator)

                self.num_of_custom_sets = len(set(gene_sets_names))
                gene_names, _ = self.custom_data.get_column_view(self.custom_gene_id_column)
                hierarchy_title = (self.custom_data.name if self.custom_data.name else 'Custom sets', )
                try:
                    self.gs_widget.add_custom_sets(
                        gene_sets_names, gene_names,
                        hierarchy_title=hierarchy_title, select_customs_flag=select_customs_flag)
                except geneset.GeneSetException:
                    pass
                # self.gs_label_combobox.setDisabled(False)
            else:
                self.gs_widget.update_gs_hierarchy()

        self.update_info_box()

    def update_tree_view(self):
        self.init_gene_sets()

    def invalidate(self):
        # clear
        self.__reset_widget_state()
        self.update_info_box()

        if self.input_data is not None:
            # setup
            self.__get_input_genes()
            self.update_tree_view()

    def __check_organism_mismatch(self):
        """ Check if organisms from different inputs match.

        :return: True if there is a mismatch
        """
        if self.tax_id is not None and self.custom_tax_id is not None:
            return self.tax_id != self.custom_tax_id
        return False

    def __get_reference_genes(self):
        self.reference_genes = []

        if self.reference_attr_names:
            for variable in self.reference_data.domain.attributes:
                self.reference_genes.append(str(variable.attributes.get(self.reference_gene_id_attribute, '?')))
        else:
            genes, _ = self.reference_data.get_column_view(self.reference_gene_id_column)
            self.reference_genes = [str(g) for g in genes]

    @Inputs.custom_sets
    def handle_custom_input(self, data):
        self.Error.clear()
        self.__reset_widget_state()
        self.custom_data = None
        self.custom_tax_id = None
        self.custom_use_attr_names = None
        self.custom_gene_id_attribute = None
        self.custom_gene_id_column = None
        self.feature_model.set_domain(None)

        if data:
            self.custom_data = data
            self.feature_model.set_domain(self.custom_data.domain)
            self.custom_tax_id = str(self.custom_data.attributes.get(TAX_ID, None))
            self.custom_use_attr_names = self.custom_data.attributes.get(GENE_AS_ATTRIBUTE_NAME, None)
            self.custom_gene_id_attribute = self.custom_data.attributes.get(GENE_ID_ATTRIBUTE, None)
            self.custom_gene_id_column = self.custom_data.attributes.get(GENE_ID_COLUMN, None)

            if self.gs_label_combobox is None:
                self.gs_label_combobox = comboBox(self.custom_gs_col_box, self, "custom_gene_set_indicator",
                                                  sendSelectedValue=True, model=self.feature_model,
                                                  callback=self.on_gene_set_indicator_changed)
            self.custom_gs_col_box.show()

            if self.custom_gene_set_indicator in self.feature_model:
                index = self.feature_model.indexOf(self.custom_gene_set_indicator)
                self.custom_gene_set_indicator = self.feature_model[index]
            else:
                self.custom_gene_set_indicator = self.feature_model[0]
        else:
            self.custom_gs_col_box.hide()

        self.gs_widget.clear_custom_sets()
        self.handle_custom_gene_sets(select_customs_flag=self.custom_gene_set_indicator is not None)
        self.invalidate()

    @Inputs.genes
    def handle_genes_input(self, data):
        self.Error.clear()
        self.__reset_widget_state()
        # clear output
        self.Outputs.matched_genes.send(None)
        # clear input values
        self.input_genes = []
        self.input_data = None
        self.tax_id = None
        self.use_attr_names = None
        self.gene_id_attribute = None
        self.gs_widget.clear()
        self.gs_widget.clear_gene_sets()
        self.update_info_box()

        if data:
            self.input_data = data
            self.tax_id = str(self.input_data.attributes.get(TAX_ID, None))
            self.use_attr_names = self.input_data.attributes.get(GENE_AS_ATTRIBUTE_NAME, None)
            self.gene_id_attribute = self.input_data.attributes.get(GENE_ID_ATTRIBUTE, None)
            self.gene_id_column = self.input_data.attributes.get(GENE_ID_COLUMN, None)
            self.update_info_box()

            if not(self.use_attr_names is not None
                   and ((self.gene_id_attribute is None) ^ (self.gene_id_column is None))):

                if self.tax_id is None:
                    self.Error.missing_annotation()
                    return

                self.Error.missing_gene_id()
                return

            elif self.tax_id is None:
                self.Error.missing_tax_id()
                return

            if self.__check_organism_mismatch():
                self.Error.organism_mismatch()
                return

            self.gs_widget.load_gene_sets(self.tax_id)

            # if input data change, we need to refresh custom sets
            if self.custom_data:
                self.gs_widget.clear_custom_sets()
                self.handle_custom_gene_sets()

            self.invalidate()

    def update_info_box(self):
        info_string = ''
        if self.input_genes:
            info_string += '{} unique gene names on input.\n'.format(len(self.input_genes))
            info_string += '{} genes on output.\n'.format(self.num_of_sel_genes)
        else:
            if self.input_data:
                if not any([self.gene_id_column, self.gene_id_attribute]):
                    info_string += 'Input data with incorrect meta data.\nUse Gene Name Matcher widget.'
            else:
                info_string += 'No data on input.\n'

        if self.custom_data:
            info_string += '{} marker genes in {} sets\n'.format(self.custom_data.X.shape[0], self.num_of_custom_sets)

        self.input_info.setText(info_string)

    def create_partial(self):
        return partial(self.set_items,
                       self.gs_widget.gs_object,
                       self.stored_gene_sets_selection,
                       set(self.input_genes),
                       self.callback)

    def callback(self):
        if self._task.cancelled:
            raise KeyboardInterrupt()
        if self.progress_bar:
            methodinvoke(self, "progress_advance")()

    def init_gene_sets(self):
        if self._task is not None:
            self.cancel()
        assert self._task is None

        self._task = Task()
        self.init_item_model()

        # save setting on selected hierarchies
        self.stored_gene_sets_selection = self.gs_widget.get_hierarchies(only_selected=True)

        f = self.create_partial()

        progress_iterations = sum([len(g_set) for hier, g_set
                                   in self.gs_widget.gs_object.map_hierarchy_to_sets().items()
                                   if hier in self.stored_gene_sets_selection])

        self.progress_bar = ProgressBar(self, iterations=progress_iterations)

        self._task.future = self._executor.submit(f)

        self._task.watcher = FutureWatcher(self._task.future)
        self._task.watcher.done.connect(self._init_gene_sets_finished)

    @Slot(concurrent.futures.Future)
    def _init_gene_sets_finished(self, f):
        assert self.thread() is QThread.currentThread()
        assert threading.current_thread() == threading.main_thread()
        assert self._task is not None
        assert self._task.future is f
        assert f.done()

        self._task = None
        self.progress_bar.finish()
        self.setStatusMessage('')

        try:
            results = f.result()  # type: list
            [self.data_model.appendRow(model_item) for model_item in results]
            self.filter_proxy_model.setSourceModel(self.data_model)
            self.data_view.selectionModel().selectionChanged.connect(self.commit)
            self.filter_data_view()
            self.set_selection()
            self.update_info_box()
        except Exception as ex:
            print(ex)

    def create_filters(self):
        search_term = self.search_pattern.lower().strip().split()

        filters = [
            FilterProxyModel.Filter(
                self.TERM, Qt.DisplayRole,
                lambda value: all(fs in value.lower() for fs in search_term))
        ]

        if self.use_min_count:
            filters.append(
                FilterProxyModel.Filter(
                    self.COUNT, Qt.DisplayRole,
                    lambda value: value >= self.min_count,
                )
            )

        return filters

    def filter_data_view(self):
        filter_proxy = self.filter_proxy_model  # type: FilterProxyModel
        model = filter_proxy.sourceModel()      # type: QStandardItemModel

        if isinstance(model, QStandardItemModel):

            # apply filtering rules
            filter_proxy.set_filters(self.create_filters())

            if model.rowCount() and not filter_proxy.rowCount():
                self.Warning.all_sets_filtered()
            else:
                self.Warning.clear()

    def set_selection(self):
        if len(self.selected_rows):
            view = self.data_view
            model = self.data_model

            row_model_indexes = [model.indexFromItem(model.item(i)) for i in self.selected_rows]
            proxy_rows = [self.filter_proxy_model.mapFromSource(i).row() for i in row_model_indexes]

            if model.rowCount() <= self.selected_rows[-1]:
                return

            header_count = view.header().count() - 1
            selection = QItemSelection()

            for row_index in proxy_rows:
                selection.append(
                    QItemSelectionRange(
                        self.filter_proxy_model.index(row_index, 0),
                        self.filter_proxy_model.index(row_index, header_count)
                    )
                )

            view.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)

    def commit(self):
        selection_model = self.data_view.selectionModel()

        if selection_model:
            selection = selection_model.selectedRows(self.COUNT)
            self.selected_rows = [self.filter_proxy_model.mapToSource(sel).row() for sel in selection]

            if selection and self.input_genes:
                genes = [model_index.data(Qt.UserRole) for model_index in selection]
                output_genes = [gene_name for gene_name in list(set.union(*genes))]
                self.num_of_sel_genes = len(output_genes)
                self.update_info_box()

                if self.use_attr_names:
                    selected = [column for column in self.input_data.domain.attributes
                                if self.gene_id_attribute in column.attributes and
                                str(column.attributes[self.gene_id_attribute]) in output_genes]

                    domain = Domain(selected, self.input_data.domain.class_vars, self.input_data.domain.metas)
                    new_data = self.input_data.from_table(domain, self.input_data)
                    self.Outputs.matched_genes.send(new_data)

                else:
                    # create filter from selected column for genes
                    only_known = table_filter.FilterStringList(self.gene_id_column, output_genes)
                    # apply filter to the data
                    data_table = table_filter.Values([only_known])(self.input_data)

                    self.Outputs.matched_genes.send(data_table)

    def assign_delegates(self):
        self.data_view.setItemDelegateForColumn(
            self.GENES, NumericalColumnDelegate(self)
        )

        self.data_view.setItemDelegateForColumn(
            self.COUNT, NumericalColumnDelegate(self)
        )

    def setup_filter_model(self):
        self.filter_proxy_model = FilterProxyModel()
        self.filter_proxy_model.setFilterKeyColumn(self.TERM)
        self.data_view.setModel(self.filter_proxy_model)

    def setup_filter_area(self):
        h_layout = QHBoxLayout()
        h_layout.setSpacing(100)
        h_widget = widgetBox(self.mainArea, orientation=h_layout)

        spin(h_widget, self, 'min_count', 0, 1000,
             label='Count',
             tooltip='Minimum genes count',
             checked='use_min_count',
             callback=self.filter_data_view,
             callbackOnReturn=True,
             checkCallback=self.filter_data_view)

        self.line_edit_filter = lineEdit(h_widget, self, 'search_pattern')
        self.line_edit_filter.setPlaceholderText('Filter gene sets ...')
        self.line_edit_filter.textChanged.connect(self.filter_data_view)

    def on_gene_set_indicator_changed(self):
        # self._handle_future_model()
        self.gs_widget.clear_custom_sets()
        self.handle_custom_gene_sets()
        self.invalidate()

    def setup_control_area(self):
        # Control area
        self.input_info = widgetLabel(
            widgetBox(self.controlArea, "Info", addSpace=True), 'No data on input.\n'
        )
        self.custom_gs_col_box = box = vBox(self.controlArea, 'Custom Gene Set Term Column')
        box.hide()

        gene_sets_box = widgetBox(self.controlArea, "Gene Sets")
        self.gs_widget = GeneSetsSelection(gene_sets_box, self, 'stored_gene_sets_selection')
        self.gs_widget.hierarchy_tree_widget.itemClicked.connect(self.update_tree_view)

        self.commit_button = auto_commit(self.controlArea, self, "auto_commit", "&Commit", box=False)

    def setup_gui(self):
        # control area
        self.setup_control_area()

        # main area
        self.data_view = QTreeView()
        self.setup_filter_model()
        self.setup_filter_area()
        self.data_view.setAlternatingRowColors(True)
        self.data_view.sortByColumn(self.COUNT, Qt.DescendingOrder)
        self.data_view.setSortingEnabled(True)
        self.data_view.setSelectionMode(QTreeView.ExtendedSelection)
        self.data_view.setEditTriggers(QTreeView.NoEditTriggers)
        self.data_view.viewport().setMouseTracking(False)
        self.data_view.setItemDelegateForColumn(self.TERM, LinkStyledItemDelegate(self.data_view))

        self.mainArea.layout().addWidget(self.data_view)

        self.data_view.header().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.assign_delegates()

    @staticmethod
    def set_items(gene_sets, sets_to_display, genes, callback):
        model_items = []
        if not genes:
            return

        for gene_set in sorted(gene_sets):
            if gene_set.hierarchy not in sets_to_display:
                continue

            callback()

            matched_set = gene_set.genes & genes
            if len(matched_set) > 0:
                category_column = QStandardItem()
                term_column = QStandardItem()
                count_column = QStandardItem()
                genes_column = QStandardItem()

                category_column.setData(", ".join(gene_set.hierarchy), Qt.DisplayRole)
                term_column.setData(gene_set.name, Qt.DisplayRole)
                term_column.setData(gene_set.name, Qt.ToolTipRole)
                term_column.setData(gene_set.link, LinkRole)
                term_column.setForeground(QColor(Qt.blue))

                count_column.setData(matched_set, Qt.UserRole)
                count_column.setData(len(matched_set), Qt.DisplayRole)

                genes_column.setData(len(gene_set.genes), Qt.DisplayRole)
                genes_column.setData(set(gene_set.genes), Qt.UserRole)  # store genes to get then on output on selection

                model_items.append([count_column, genes_column, term_column, category_column])

        return model_items

    def init_item_model(self):
        if self.data_model:
            self.data_model.clear()
            self.setup_filter_model()
        else:
            self.data_model = QStandardItemModel()

        self.data_model.setSortRole(Qt.UserRole)
        self.data_model.setHorizontalHeaderLabels(self.DATA_HEADER_LABELS)

    def sizeHint(self):
        return QSize(1280, 960)


if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication
    app = QApplication([])
    ow = OWGeneSets()
    ow.show()
    app.exec_()



""" GeneSet module """
from orangecontrib.bioinformatics.geneset.config import *
from orangecontrib.bioinformatics.geneset.utils import (
    filename, GeneSets, GeneSet, NoGeneSetsException, GeneSetException, filename_parse
)
from orangecontrib.bioinformatics.utils import serverfiles

from typing import Tuple


def list_all(**kwargs):
    """ Returns available gene sets from the server files repository.

    :param kwargs:
        * *organism* (``str``) -- Taxonomy id (NCBI taxonomy database)


    :rtype: :obj:`list` of (hierarchy, organism)


    Example
    --------
    The available gene set collection can be listed with:
        >>> list_all(organism='10090')

    """

    organism = kwargs.get("organism", None)

    all_available = set([filename_parse(f_name) for domain, f_name
                         in serverfiles.ServerFiles().listfiles(DOMAIN) + serverfiles.listfiles(DOMAIN)])
    if organism:
        return [hier for hier, org in all_available if org == organism]
    else:
        return all_available

      
def load_gene_sets(hierarchy, tax_id):
      # type: (Tuple[Tuple(str, str), str]) -> GeneSets
    """ Initialize gene sets from a given hierarchy.

    :param tuple hierarchy: gene set hierarchy.
    :rtype: :obj:`GeneSets`

    Example
    --------
    Gene sets provided with Orange are organized hierarchically:
        >>> list_of_genesets= list_all(organism='10090')
            [(('KEGG', 'Pathways'), '10090'),
             (('KEGG', 'pathways'), '10090'),
             (('GO', 'biological_process'), '10090'),
             (('GO', 'molecular_function'), '10090'),
             (('GO', 'cellular_component'), '10090')]
        >>> load_gene_sets(list_of_genesets[0])

    """
    file_path = serverfiles.localpath_download(DOMAIN, filename(hierarchy, tax_id))
    return GeneSets.from_gmt_file_format(file_path)



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring
import unittest
from unittest.mock import Mock

import numpy as np

from Orange.data import Table, Variable
from Orange.data.filter import FilterString, Values
from Orange.projection import PCA, MDS
from Orange.widgets.tests.base import WidgetTest, WidgetOutputsTestMixin, \
    ProjectionWidgetTestMixin, simulate

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.widgets.OWAnnotateProjection import \
    OWAnnotateProjection
from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID


class TestOWAnnotateProjection(WidgetTest, ProjectionWidgetTestMixin,
                               WidgetOutputsTestMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        path = "https://datasets.orange.biolab.si/sc/aml-1k.tab.gz"
        Variable._clear_all_caches()
        cls.data = Table(path)[::50]
        cls.data.attributes[TAX_ID] = "9606"

        cls.signal_name = "Data"
        cls.signal_data = cls.data
        cls.same_input_output_domain = False

        genes_path = serverfiles.localpath_download(
            "marker_genes", "panglao_gene_markers.tab")
        f = FilterString("Organism", FilterString.Equal, "Human")
        cls.genes = Values([f])(Table(genes_path))

    def setUp(self):
        self.widget = self.create_widget(OWAnnotateProjection)
        self.send_signal(self.widget.Inputs.projector, PCA())

    def tearDown(self):
        self.widget.cancel()
        self.wait_until_stop_blocking()

    def test_input_projector(self):
        self.send_signal(self.widget.Inputs.projector, None)
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        self.assertFalse(self.widget.Error.proj_error.is_shown())
        self.assertIsNotNone(self.widget.embedding)
        self.send_signal(self.widget.Inputs.projector, PCA())
        self.assertIsInstance(self.widget.projector, PCA)
        self.wait_until_stop_blocking()
        self.send_signal(self.widget.Inputs.projector, MDS())
        self.wait_until_stop_blocking()
        self.assertTrue(self.widget.Error.proj_error.is_shown())

    def test_input_genes(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertTrue(self.widget.Warning.no_genes.is_shown())
        self.wait_until_stop_blocking()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.assertFalse(self.widget.Warning.no_genes.is_shown())
        self.wait_until_stop_blocking()
        self.send_signal(self.widget.Inputs.genes, None)
        self.assertTrue(self.widget.Warning.no_genes.is_shown())
        self.wait_until_stop_blocking()
        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(self.widget.Warning.no_genes.is_shown())

    @unittest.skip("Skip due to timeout")
    def test_scoring_method_control(self):
        def start():
            self.widget.run_button.click()
            self.wait_until_stop_blocking()

        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        output1 = self.get_output(self.widget.Outputs.annotated_data)
        cbox = self.widget.controls.scoring_method
        simulate.combobox_activate_index(cbox, 2)
        start()
        output2 = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(output1.X, output2.X)
        np.testing.assert_array_equal(output1.Y, output2.Y)

    @unittest.skip("Skip due to timeout")
    def test_statistical_test_control(self):
        def start():
            self.widget.run_button.click()
            self.wait_until_stop_blocking()

        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        output1 = self.get_output(self.widget.Outputs.annotated_data)
        cbox = self.widget.controls.statistical_test
        simulate.combobox_activate_index(cbox, 1)
        start()
        output2 = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(output1.X, output2.X)
        np.testing.assert_array_equal(output1.Y, output2.Y)
        self.assertFalse((output1.metas == output2.metas).all())

    def test_p_threshold_control(self):
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        output1 = self.get_output(self.widget.Outputs.annotated_data)
        self.widget.controls.p_threshold.valueChanged.emit(0.1)
        self.widget.run_button.click()
        self.wait_until_stop_blocking()
        output2 = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(output1.X, output2.X)
        np.testing.assert_array_equal(output1.Y, output2.Y)

    def test_epsilon_control(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertFalse(self.widget.epsilon_spin.isEnabled())
        self.widget.controls.use_user_epsilon.click()
        self.assertTrue(self.widget.epsilon_spin.isEnabled())
        self.assertFalse(self.widget.Information.modified.is_shown())
        self.widget.controls.use_user_epsilon.click()
        self.assertTrue(self.widget.Information.modified.is_shown())

    def test_output_data(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.wait_until_stop_blocking()
        output = self.get_output(self.widget.Outputs.annotated_data)
        n = len(self.data.domain.metas)
        self.assertGreater(len(output.domain.metas), n + 4)
        self.assertListEqual(
            ["PC1", "PC2", "Clusters", "Annotation"] +
            [m.name for m in self.data.domain.metas] + ["Selected"],
            [m.name for m in output.domain.metas[-n-5:]])
        np.testing.assert_array_equal(output.X, self.data.X)
        np.testing.assert_array_equal(output.Y, self.data.Y)
        np.testing.assert_array_equal(output.metas[:, -n - 1:-1],
                                      self.data.metas)

    def test_button_no_data(self):
        self.widget.run_button.click()
        self.assertEqual(self.widget.run_button.text(), "Start")

    def test_button_with_data(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertEqual(self.widget.run_button.text(), "Stop")
        self.wait_until_stop_blocking()
        self.assertEqual(self.widget.run_button.text(), "Start")

    def test_button_toggle(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.run_button.click()
        self.assertEqual(self.widget.run_button.text(), "Resume")

    def test_plot_once(self):
        self.widget.setup_plot = Mock()
        self.widget.commit = Mock()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.widget.commit.assert_called_once()
        self.widget.commit.reset_mock()
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.setup_plot.assert_called_once()
        self.widget.commit.assert_called_once()
        self.wait_until_stop_blocking()
        self.widget.setup_plot.reset_mock()
        self.widget.commit.reset_mock()
        self.send_signal(self.widget.Inputs.data_subset, self.data[::10])
        self.widget.setup_plot.assert_not_called()
        self.widget.commit.assert_called_once()

    def test_saved_selection(self):
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()

        self.widget.graph.select_by_indices(list(range(0, len(self.data), 10)))
        settings = self.widget.settingsHandler.pack_data(self.widget)
        w = self.create_widget(self.widget.__class__, stored_settings=settings)

        self.send_signal(w.Inputs.genes, self.genes)
        self.send_signal(w.Inputs.data, self.data, widget=w)
        self.wait_until_stop_blocking(widget=w)

        self.assertEqual(np.sum(w.graph.selection), 2)
        np.testing.assert_equal(self.widget.graph.selection, w.graph.selection)

    def test_outputs(self):
        self.send_signal(self.widget.Inputs.genes, self.genes)
        super().test_outputs()

    def test_color_by_cluster(self):
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        self.assertTrue(self.widget.controls.attr_color.isEnabled())

        simulate.combobox_activate_index(self.widget.controls.attr_color, 0)
        is_grey = [brush.color().name() == "#808080" for brush in
                   self.widget.graph.scatterplot_item.data['brush']]
        self.assertTrue(all(is_grey))
        self.widget.controls.color_by_cluster.click()
        self.assertFalse(self.widget.controls.attr_color.isEnabled())
        is_not_grey = [brush.color().name() != "#808080" for brush in
                       self.widget.graph.scatterplot_item.data['brush']]
        self.assertTrue(any(is_not_grey))

    def test_attr_label_metas(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        simulate.combobox_activate_item(self.widget.controls.attr_label,
                                        self.data.domain[-1].name)

    def test_attr_models(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        controls = self.widget.controls
        self.assertEqual(len(controls.attr_color.model()), 1006)
        self.assertEqual(len(controls.attr_shape.model()), 5)
        self.assertEqual(len(controls.attr_size.model()), 1002)
        self.assertEqual(len(controls.attr_label.model()), 1008)
        for var in self.data.domain.variables + self.data.domain.metas:
            self.assertIn(var, controls.attr_label.model())
            if var.is_continuous:
                self.assertIn(var, controls.attr_color.model())
                self.assertIn(var, controls.attr_size.model())
                self.assertNotIn(var, controls.attr_shape.model())
            if var.is_discrete:
                self.assertIn(var, controls.attr_color.model())
                self.assertNotIn(var, controls.attr_size.model())
                self.assertIn(var, controls.attr_shape.model())

    def test_subset_data_color(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        self.send_signal(self.widget.Inputs.data_subset, self.data[:10])
        subset = [brush.color().name() != "#000000" for brush in
                  self.widget.graph.scatterplot_item.data['brush'][:10]]
        other = [brush.color().name() == "#000000" for brush in
                 self.widget.graph.scatterplot_item.data['brush'][10:]]
        self.assertTrue(all(subset))
        self.assertTrue(all(other))


if __name__ == "__main__":
    unittest.main()



from unittest import TestCase
from orangecontrib.bioinformatics.dicty import phenotypes as DictyMutants


class DictyMutantsTest(TestCase):

    def test_size(self):
        self.assertGreater(len(DictyMutants.gene_mutants()), 1000)



""" Cluster analysis module """
import numpy as np
import threading
import concurrent.futures

from typing import Union
from operator import attrgetter
from functools import partial


from AnyQt.QtCore import (
    Qt, QAbstractListModel, QVariant, Slot, QThread
)
from Orange.widgets.utils.concurrent import ThreadExecutor, FutureWatcher, methodinvoke
from Orange.widgets.gui import ProgressBar

from orangecontrib.bioinformatics.geneset import GeneSet
from orangecontrib.bioinformatics.widgets.utils.gui import gene_scoring_method
from orangecontrib.bioinformatics.utils.statistics import FDR, ALT_GREATER
from orangecontrib.bioinformatics.ncbi.gene import Gene

DISPLAY_GENE_COUNT = 20
DISPLAY_GENE_SETS_COUNT = 5


class ClusterGene(Gene):
    __slots__ = ['score', 'p_val', 'fdr']

    def __init__(self, gene_name, gene_id):
        super().__init__()
        self.input_name = gene_name
        self.ncbi_id = gene_id

        # default values
        self.fdr = 1
        self.p_val = 1
        self.score = 0


class ClusterGeneSet(GeneSet):
    __slots__ = ['p_val', 'fdr', 'count']

    def __init__(self):
        super().__init__()


class Cluster:

    CLUSTER_VS_REST = False
    CLUSTER_VS_CLUSTER = True

    def __init__(self, name, index):
        # type: (str, int) -> None
        self.name = name
        self.index = index

        # gene enrichment
        self.genes = []
        self.filtered_genes = []

        # set enrichment
        self.gene_sets = []
        self.filtered_gene_sets = []

        # Statistical method used when performing analysis
        self.method_used = None

    def set_genes(self, gene_names, gene_ids):
        self.genes = []

        for name, entrez in zip(gene_names, gene_ids):
            self.genes.append(ClusterGene(name, entrez))

    @staticmethod
    def apply_filter(obj, p_val, fdr):
        filter_status = []

        if p_val is not None:
            filter_status.append(obj.p_val < p_val)

        if fdr is not None:
            filter_status.append(obj.fdr < fdr)

        return all(filter_status)

    def filter_enriched_genes(self, p_val, fdr, max_gene_count=None):
        all_args_none = all(arg is None for arg in [p_val, fdr])
        filter_function = partial(self.apply_filter, p_val=p_val, fdr=fdr)
        sorted_list = sorted(self.genes, key=attrgetter('p_val' if all_args_none else 'fdr'))
        filtered_list = list(filter(filter_function, sorted_list))

        if max_gene_count is not None:
            self.filtered_genes = filtered_list[:max_gene_count]
        else:
            self.filtered_genes = filtered_list

    def filter_gene_sets(self, p_val, fdr, max_set_count=None):
        all_args_none = all(arg is None for arg in [p_val, fdr])
        filter_function = partial(self.apply_filter, p_val=p_val, fdr=fdr)
        sorted_list = sorted(self.gene_sets, key=attrgetter('p_val' if all_args_none else 'fdr'))
        filtered_list = list(filter(filter_function, sorted_list))

        if max_set_count is not None:
            self.filtered_gene_sets = filtered_list[:max_set_count]
        else:
            self.filtered_gene_sets = filtered_list

    def gene_set_enrichment(self, gene_sets, selected_sets, genes, ref_genes):
        self.gene_sets = []

        if not genes:
            return

        # calculate gene set enrichment
        for gene_set in gene_sets:

            if gene_set.hierarchy not in selected_sets:
                continue

            gs = ClusterGeneSet()
            enrichment_result = gene_set.set_enrichment(ref_genes, genes.intersection(genes))
            gs.count = len(enrichment_result.query)
            gs.p_val = enrichment_result.p_value
            gs.name = gene_set.name
            gs.gs_id = gene_set.gs_id
            self.gene_sets.append(gs)

        # calculate FDR
        p_vals = [gs.p_val for gs in self.gene_sets]
        fdr_values = FDR(p_vals)

        for gs, fdr in zip(self.gene_sets, fdr_values):
            gs.fdr = fdr

    def __update_gene_objects(self, scores, p_vals, fdr_vals):
        # type: (Union[np.ndarray, list], Union[np.ndarray, list], Union[np.ndarray, list]) ->  None
        """ update gene objects with computed results

        :param p_vals:   Computed p-values
        :param fdr_vals: Computed fdr-values

        """
        for score, p_val, fdr, gene in zip(scores, p_vals, fdr_vals, self.genes):
            gene.score = score
            gene.p_val = p_val
            gene.fdr = fdr

    def cluster_scores(self, table_x, rows_by_cluster, method, design, **kwargs):
        # type: (np.ndarray, np.ndarray, gene_scoring_method, str) -> None
        """
        General scoring of genes in the cluster.
        Ways to score genes are determined by design.
        If a batch variable index is defined, it is accounted for in gene scoring.

        :param table_x:
        :param rows_by_cluster:
        :param method:
        :param design:
        :param kwargs:
        :param aggregation:
        :param rows_by_batch:
        :return:
        """
        aggregation = kwargs.get('aggregation', 'max')
        alternative = kwargs.get('alternative', ALT_GREATER)
        rows_by_batch = kwargs.get('rows_by_batch', None)
        if not isinstance(rows_by_batch, np.ndarray):
            rows_by_batch = np.zeros((len(table_x),))
        uniq_batches = set(rows_by_batch)

        # Determine clusters
        self.method_used = method.name
        uniq_clusters = set(rows_by_cluster) - {self.index}
        this_cluster = self.index
        if design == self.CLUSTER_VS_REST:
            rows_by_cluster = rows_by_cluster == self.index
            uniq_clusters = {False}
            this_cluster = True

        calculated_p_values = np.ones((table_x.shape[1],      # genes
                                      len(uniq_clusters),     # other clusters
                                      len(uniq_batches)))     # batches

        calculated_scores = np.ones((table_x.shape[1],       # genes
                                     len(uniq_clusters),     # other clusters
                                     len(uniq_batches)))     # batches

        for bi, b in enumerate(uniq_batches):
            for ci, c in enumerate(uniq_clusters):
                cluster = table_x[np.logical_and(rows_by_cluster == this_cluster, rows_by_batch == b)]
                rest = table_x[np.logical_and(rows_by_cluster == c, rows_by_batch == b)]
                if cluster.any() and rest.any():
                    scores, p_values = method.score_function(cluster, rest, alternative=alternative)
                    scores[np.isnan(p_values)] = 0
                    calculated_scores[:, ci, bi] = scores
                    p_values[np.isnan(p_values)] = 1
                    calculated_p_values[:, ci, bi] = p_values

        if aggregation == 'max':
            max_p_values = np.max(calculated_p_values, axis=(1, 2))
            max_p_indexes = np.where(np.max(calculated_p_values, axis=(1, 2), keepdims=True) == calculated_p_values)
            # this holds true only if max_p_indexes.ndim == 3
            scores = calculated_scores[max_p_indexes[0], max_p_indexes[1], max_p_indexes[2]]
            fdr_values = FDR(max_p_values.tolist())
            self.__update_gene_objects(scores, max_p_values, fdr_values)
        else:
            raise NotImplementedError("Aggregation %s is not implemented" % aggregation)
        return

    def to_html(self):
        gene_sets = '(no enriched gene sets)'
        if self.filtered_gene_sets:
            sets_to_display = self.filtered_gene_sets[:DISPLAY_GENE_SETS_COUNT]
            gene_sets = '<br>'.join(['<b>{}</b> (FDR={:0.2e}, n={})'.format(g_set.name, g_set.fdr, g_set.count)
                                    for g_set in sets_to_display])

            if len(self.filtered_gene_sets) > len(sets_to_display):
                gene_sets += \
                    '<br> ... ({} more gene sets)'.format(len(self.filtered_gene_sets) - DISPLAY_GENE_SETS_COUNT)

        genes = '(all genes are filtered out)'
        if self.filtered_genes:
            genes_to_display = [gene.input_name for gene in self.filtered_genes[:DISPLAY_GENE_COUNT]]

            genes = ', '.join(genes_to_display)
            if len(self.filtered_genes) > len(genes_to_display):
                genes += ', ... ({} more genes)'.format(len(self.filtered_genes) - DISPLAY_GENE_COUNT)

        html_string = """
        <html>
        <body>
        <table >
            <tr>
            </tr>
            <tr>
                <td width=\"20%\" heigth=\"100%\" align=left valign=top><p>{}</p></td>
                <td width=\"80%\" heigth=\"100%\" align=left valign=top>
                    {}
                </td>
            </tr>
            <tr> </tr>
            <tr>
                <td width=\"20%\" ></td>
                <td width=\"80%\" align=left valign=top>{}</td>
            </tr>
            <tr>
            </tr>
        </table>
        </body>
        </html>
        """.format(self.name, gene_sets, genes)
        return html_string


class Task:
    future = None
    watcher = None
    cancelled = False

    def cancel(self):
        self.cancelled = True
        self.future.cancel()
        concurrent.futures.wait([self.future])


class ClusterModel(QAbstractListModel):

    def __init__(self, parent=None):
        QAbstractListModel.__init__(self)
        self.__items = []
        self.parent = parent

        self._task = None  # type: Union[Task, None]
        self._executor = ThreadExecutor()

    def add_rows(self, rows):
        self.__items = rows

    def get_rows(self):
        return self.__items

    def rowCount(self, *args, **kwargs):
        return len(self.__items)

    def data(self, model_index, role=None):
        # check if data is set
        if not self.__items:
            return QVariant()

        # return empty QVariant if model index is unknown
        if not model_index.isValid() or not (0 <= model_index.row() < len(self.__items)):
            return QVariant()

        row_obj = self.__items[model_index.row()]

        if role == Qt.DisplayRole:
            return row_obj

    @Slot(concurrent.futures.Future)
    def _end_task(self, f):
        assert self.thread() is QThread.currentThread()
        assert threading.current_thread() == threading.main_thread()
        assert self._task is not None
        assert self._task.future is f
        assert f.done()

        self._task = None
        self.parent.progressBarFinished()
        self.parent.filter_genes()

        try:
            f.result()
        except Exception as ex:
            raise ex

    def _score_genes(self, callback, **kwargs):
        for item in self.get_rows():
            item.cluster_scores(**kwargs)
            callback()

    @Slot(bool)
    def progress_advance(self, finish):
        # GUI should be updated in main thread. That's why wex are calling advance method here
        if self.parent.progress_bar:
            if finish:
                self.parent.progressBarFinished()
            else:
                self.parent.progress_bar.advance()

    def cancel(self):
        """
        Cancel the current task (if any).
        """
        if self._task is not None:
            self._task.cancel()
            assert self._task.future.done()
            # disconnect the `_task_finished` slot
            self._task.watcher.done.disconnect(self._end_task)
            self._task = None

    def score_genes(self, **kwargs):
        """ Run gene enrichment.

        :param design:
        :param data_x:
        :param rows_by_cluster:
        :param method:


        Note:
            We do not apply filter nor notify view that data is changed. This is done after filters
        """
        if self._task is not None:
            # First make sure any pending tasks are cancelled.
            self.cancel()
        assert self._task is None

        progress_advance = methodinvoke(self, "progress_advance", (bool,))

        def callback():
            if self._task.cancelled:
                raise KeyboardInterrupt()
            progress_advance(self._task.cancelled)

        self.parent.progress_bar = ProgressBar(self.parent, iterations=len(self.get_rows()))
        f = partial(self._score_genes, callback=callback, **kwargs)
        self._task = Task()
        self._task.future = self._executor.submit(f)

        self._task.watcher = FutureWatcher(self._task.future)
        self._task.watcher.done.connect(self._end_task)

    def gene_sets_enrichment(self, gs_object, gene_sets, reference_genes):
        """ Run gene sets enrichment.

        :param gs_object:
        :param gene_sets:
        :param reference_genes:

        Note:
            We do not apply filter nor notify view that data is changed. This is done after filters

        """

        for item in self.get_rows():
            genes = [gene.ncbi_id for gene in item.filtered_genes]
            item.gene_set_enrichment(gs_object,
                                     gene_sets,
                                     set(genes),
                                     reference_genes)

    def apply_gene_filters(self, p_val=None, fdr=None, count=None):
        [item.filter_enriched_genes(p_val, fdr, max_gene_count=count) for item in self.get_rows()]
        self.dataChanged.emit(self.createIndex(0, 0), self.createIndex(self.rowCount(0), 0))

    def apply_gene_sets_filters(self, p_val=None, fdr=None, count=None):
        [item.filter_gene_sets(p_val, fdr, max_set_count=count) for item in self.get_rows()]
        self.dataChanged.emit(self.createIndex(0, 0), self.createIndex(self.rowCount(0), 0))







""" dictyExpress widget """
import sys
import threading


from requests.exceptions import ConnectionError

from AnyQt.QtCore import Qt, QThreadPool, QSize
from AnyQt.QtGui import QFont
from AnyQt.QtWidgets import QTreeWidget, QTreeWidgetItem, QLabel, QLineEdit, QApplication, QFrame

from Orange.data import Table, StringVariable, Domain
from Orange.widgets import gui, settings
from Orange.widgets.widget import OWWidget, Msg
from Orange.widgets.utils.signals import Output


from orangecontrib.bioinformatics import resolwe
from orangecontrib.bioinformatics.resolwe.utils import etc_to_table
from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID, GENE_AS_ATTRIBUTE_NAME, GENE_ID_ATTRIBUTE, GENE_ID_COLUMN
)
from orangecontrib.bioinformatics.ncbi.gene import NCBI_ID, GeneMatcher
from orangecontrib.bioinformatics.widgets.utils.concurrent import Worker


Labels = [
    (" ", " "),
    ("var.project", "Project"),
    ("static.name", "Experiment"),
    ("static.cite", "Citation"),
    ("var.growth", "Growth"),
    ("var.treatment", "Treatment"),
    ("var.strain", "Strain")]


# Creates line separator
def h_line():
    line = QFrame()
    line.setFrameShape(QFrame.HLine)
    line.setFrameShadow(QFrame.Sunken)
    return line


class OWdictyExpress(OWWidget):

    name = "dictyExpress"
    description = "Time-course gene expression data"
    icon = "../widgets/icons/OWdictyExpress.png"
    want_main_area = True
    priority = 3

    class Inputs:
        pass

    class Outputs:
        etc_data = Output("Data", Table)

    class Error(OWWidget.Error):
        unreachable_host = Msg('Host not reachable')
        invalid_credentials = Msg('Invalid credentials')

    username = settings.Setting('')
    # password = settings.Setting('')
    setTimeVariable = settings.Setting(False)
    gene_as_attr_name = settings.Setting(0)

    selected_item = settings.Setting(None, schema_only=True)
    auto_commit = settings.Setting(False, schema_only=True)

    def __init__(self):
        super().__init__()

        self.res = None
        self.orgnism = 352472
        self.server = 'https://dictyexpress.research.bcm.edu'
        self.headerLabels = [x[1] for x in Labels]
        self.searchString = ""
        self.items = []

        self.progress_bar = None
        # threads
        self.threadpool = QThreadPool()

        # Login Section
        box = gui.widgetBox(self.controlArea, 'Login')

        self.namefield = gui.lineEdit(box, self, "username", "Username:", labelWidth=100, orientation='horizontal',
                                      callback=self.auth_changed)

        self.password = ''
        self.passfield = gui.lineEdit(box, self, "password", "Password:", labelWidth=100, orientation='horizontal',
                                      callback=self.auth_changed)

        self.passfield.setEchoMode(QLineEdit.Password)

        self.controlArea.layout().addWidget(h_line())

        box = gui.widgetBox(self.controlArea, "Output", addSpace=True)
        gui.radioButtonsInBox(box, self, "gene_as_attr_name", ["Genes in rows", "Genes in columns"],
                              callback=self.invalidate)

        self.controlArea.layout().addWidget(h_line())

        self.refresh_button = gui.button(self.controlArea, self, "Refresh", callback=self.refresh)
        self.handle_cache_button(True)

        gui.rubber(self.controlArea)

        self.commit_button = gui.auto_commit(self.controlArea, self, "auto_commit", "&Commit", box=False)

        # Experiment Section

        label = QLabel("Available projects:")
        my_font = QFont()
        my_font.setBold(True)
        label.setFont(my_font)
        self.mainArea.layout().addWidget(label)

        self.mainArea.layout().addWidget(h_line())

        self.filter = gui.lineEdit(self.mainArea, self, "searchString", "Filter:", callbackOnType=True,
                                   callback=self.search_update)

        self.experimentsWidget = QTreeWidget(alternatingRowColors=True,
                                             rootIsDecorated=False,
                                             uniformRowHeights=True,
                                             sortingEnabled=True)

        self.experimentsWidget.setItemDelegateForColumn(
            0, gui.IndicatorItemDelegate(self, role=Qt.DisplayRole))

        self.experimentsWidget.selectionModel().selectionChanged.connect(
            self.onSelectionChanged)

        self.experimentsWidget.setHeaderLabels(self.headerLabels)
        self.mainArea.layout().addWidget(self.experimentsWidget)

        self.auth_set()
        self.connect()
        self.sizeHint()

    def sizeHint(self):
        return QSize(1400, 680)

    def auth_set(self):
        self.passfield.setDisabled(not self.username)

    def auth_changed(self):
        self.auth_set()
        self.connect()

    def refresh(self):
        self.reset()
        self.load_experiments()

    def reset(self):
        self.experimentsWidget.clear()  # clear QTreeWidget
        self.items = []
        # self.lastSelected = None
        self.searchString = ""

    def search_update(self):
        parts = self.searchString.split()
        for item in self.items:
            item.setHidden(not all(s in item for s in parts))

    def progress_advance(self):
        # GUI should be updated in main thread. That's why we are calling advance method here
        assert threading.current_thread() == threading.main_thread()
        if self.progress_bar:
            self.progress_bar.advance()

    def handle_error(self, ex):
        self.progress_bar.finish()
        self.setStatusMessage('')
        if isinstance(ex, ConnectionError) or isinstance(ex, ValueError):
            self.Error.unreachable_host()

        print(ex)

    def load_experiments_result(self, experiments):
        self.load_tree_items(experiments)
        self.progress_bar.finish()
        self.setStatusMessage('')

    def connect(self):
        self.res = None
        self.Error.clear()
        self.reset()
        self.handle_cache_button(False)

        user, password = resolwe.DEFAULT_EMAIL, resolwe.DEFAULT_PASSWD
        if self.username or self.password:
            user, password = self.username, self.password

        try:
            self.res = resolwe.connect(user, password, self.server, 'genesis')
        except resolwe.ResolweAuthException as e:
            self.Error.invalid_credentials()
        else:
            self.load_experiments()
            self.handle_cache_button(True)

    def load_experiments(self):
        if self.res:
            # init progress bar
            self.progress_bar = gui.ProgressBar(self, iterations=2)
            # status message
            self.setStatusMessage('downloading experiments')

            worker = Worker(self.res.fetch_etc_objects, progress_callback=True)
            worker.signals.progress.connect(self.progress_advance)
            worker.signals.result.connect(self.load_experiments_result)
            worker.signals.error.connect(self.handle_error)

            # move download process to worker thread
            self.threadpool.start(worker)

    def load_tree_items(self, list_of_exp):
        self.items = [CustomTreeItem(self.experimentsWidget, item) for item in list_of_exp]

        for i in range(len(self.headerLabels)):
            self.experimentsWidget.resizeColumnToContents(i)

        self.set_cached_indicator()
        self.set_selected()

    def set_selected(self):
        for item in self.items:
            if self.selected_item and item.gen_data_id == self.selected_item:
                self.experimentsWidget.setCurrentItem(item)

    def onSelectionChanged(self):
        self.invalidate()

    def invalidate(self):
        self.commit()

    def handle_cache_button(self, handle):
        self.refresh_button.setEnabled(handle)

    def send_to_output(self, result):
        self.progress_bar.finish()
        self.setStatusMessage('')

        etc_json, table_name = result

        # convert to table
        data = etc_to_table(etc_json, bool(self.gene_as_attr_name))
        # set table name
        data.name = table_name

        # match genes
        gene_matcher = GeneMatcher(str(self.orgnism))

        if not bool(self.gene_as_attr_name):
            if 'Gene' in data.domain:
                gene_column = data.domain['Gene']
                gene_names = data.get_column_view(gene_column)[0]
                gene_matcher.genes = gene_names
                gene_matcher.run_matcher()

                domain_ids = Domain([], metas=[StringVariable(NCBI_ID)])
                data_ids = [[str(gene.ncbi_id) if gene.ncbi_id else '?'] for gene in gene_matcher.genes]
                table_ids = Table(domain_ids, data_ids)
                data = Table.concatenate([data, table_ids])

            data.attributes[GENE_ID_COLUMN] = NCBI_ID
        else:
            gene_matcher.match_table_attributes(data)
            data.attributes[GENE_ID_ATTRIBUTE] = NCBI_ID

        # add table attributes
        data.attributes[TAX_ID] = str(self.orgnism)
        data.attributes[GENE_AS_ATTRIBUTE_NAME] = bool(self.gene_as_attr_name)

        # reset cache indicators
        self.set_cached_indicator()
        # send data to the output signal
        self.Outputs.etc_data.send(data)

    def commit(self):
        self.Error.clear()
        selected_item = self.experimentsWidget.currentItem()  # get selected TreeItem
        self.selected_item = selected_item.gen_data_id

        if selected_item:
            # init progress bar
            self.progress_bar = gui.ProgressBar(self, iterations=1)
            # status message
            self.setStatusMessage('downloading experiment data')

            worker = Worker(self.res.download_etc_data,
                            selected_item.gen_data_id,
                            table_name=selected_item.data_name,
                            progress_callback=True)

            worker.signals.progress.connect(self.progress_advance)
            worker.signals.result.connect(self.send_to_output)
            worker.signals.error.connect(self.handle_error)

            # move download process to worker thread
            self.threadpool.start(worker)

    def set_cached_indicator(self):
        cached = self.res.get_cached_ids()

        for item in self.items:

            if item.gen_data_id in cached:
                item.setData(0, Qt.DisplayRole, " ")
            else:
                item.setData(0, Qt.DisplayRole, "")


class CustomTreeItem(QTreeWidgetItem):

    def __init__(self, parent, gen_data):
        super(CustomTreeItem, self).__init__(parent)  # Init super class (QtGui.QTreeWidgetItem )

        self._gen_data = gen_data  # GenData object
        self.set_rows(self._gen_data.annotation)  # set rows in QTreeWidget

    def __contains__(self, text):
        return any(text.upper() in str(self.text(i)).upper() for i in range(self.columnCount()))

    @property
    def gen_data_id(self):
        return self._gen_data.id

    @property
    def data_name(self):
        try:
            project = self._gen_data.var['project']
            experiment = self._gen_data.static['name']
        except AttributeError:
            project = ''
            experiment = ''

        return '{} ({})'.format(project, experiment)

    def set_rows(self, row):
        for index, label in enumerate(Labels):
            if index > 0:
                try:
                    if type(row[label[0]]["value"]) == list:
                        self.setText(index, row[label[0]]["value"][0]["name"])
                    else:
                        self.setText(index, row[label[0]]["value"])
                except IndexError:
                    self.setText(index, 'No data')


if __name__ == "__main__":

    def test_main():
        app = QApplication(sys.argv)
        w = OWdictyExpress()
        w.show()
        r = app.exec_()
        w.saveSettings()
        return r

    sys.exit(test_main())



"""common functions"""
import json
import os

from collections import OrderedDict
from datetime import datetime

INFO_FILE_SCHEMA = {
    'domain': None,
    'filename': None,
    'source': None,
    'title': None,
    'tags': [],
    'size': None,
    'datetime': None
    # used only if files are compressed
    # 'uncompressed': None,
    # 'compression': None,
}


def file_size_bytes(file_path):
    """ returns file size in bytes """
    return os.stat(file_path).st_size


def create_info_file(file_path, **kwargs):
    info_dict = OrderedDict(INFO_FILE_SCHEMA)

    info_dict.update(**kwargs)
    info_dict['datetime'] = '{0:%Y-%m-%d %H:%M:%S.%f}'.format(datetime.today())
    info_dict['size'] = file_size_bytes(file_path)

    with open(file_path + '.info', 'wt') as f:
        json.dump(info_dict, f)


def create_folder(path):
    try:
        os.makedirs(path)
    except OSError:
        if os.path.exists(path):
            #raise
            pass
        else:
            # There was an error on creation, so make sure we know about it
            raise



""" Gene Expression Omnibus datasets widget """
import sys
import numpy as np


from typing import  Optional, DefaultDict, Any
from types import SimpleNamespace as namespace
from collections import defaultdict, OrderedDict
from functools import lru_cache

from AnyQt.QtWidgets import (
    QSplitter,
    QTableView,
    QAbstractItemView,
    QStyle,
    QTreeWidget,
    QTreeWidgetItem,
    QAbstractScrollArea,
)
from AnyQt.QtCore import (
    Qt,
    QSize,
    QVariant,
    QModelIndex
)
from AnyQt.QtGui import QFont, QColor

from Orange.widgets.gui import (
    widgetLabel,
    widgetBox,
    radioButtonsInBox,
    separator,
    auto_commit,
    rubber,
    lineEdit,
    LinkRole,
    LinkStyledItemDelegate,
    IndicatorItemDelegate
)

from Orange.widgets.widget import OWWidget
from Orange.widgets.utils import itemmodels
from Orange.widgets.settings import Setting
from Orange.widgets.utils.signals import Output
from Orange.data import Table
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin

from orangecontrib.bioinformatics.geo import local_files
from orangecontrib.bioinformatics.geo.dataset import GDSInfo
from orangecontrib.bioinformatics.geo import pubmed_url, is_cached
from orangecontrib.bioinformatics.geo.dataset import get_samples, dataset_download


class Result(namespace):
    gds_dataset: Optional[Table] = None


def run_download_task(gds_id: str, samples: DefaultDict[str, list], transpose: bool, state: TaskState):
    res = Result()
    current_iter = 0
    max_iter = 102

    def callback():
        nonlocal current_iter
        current_iter += 1
        state.set_progress_value(100 * (current_iter / max_iter))

    state.set_status("Downloading...")
    res.gds_dataset = dataset_download(gds_id, samples, transpose=transpose, callback=callback)
    return res


class GEODatasetsModel(itemmodels.PyTableModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setHorizontalHeaderLabels(
            [
                '',
                'ID',
                'PubMedID',
                'Organism',
                'Samples',
                'Features',
                'Genes',
                'Subsets',
                'Title',
            ]
        )

        self.indicator_col = 0
        self.gds_id_col = 1
        self.organism_col = 3
        self.pubmedid_col = 2

        self.info = None
        self.table = None
        self.filter_mask = None

        self.font = QFont()
        self.font.setUnderline(True)
        self.color = QColor(Qt.blue)

        @lru_cache(maxsize=10000)
        def _row_instance(row, column):
            return self[int(row)][int(column)]

        self._row_instance = _row_instance

    def initialize(self, info: OrderedDict):
        self.info = info

        def __gds_to_row(gds: dict):
            gds_id = gds['name']
            title = gds['title']
            organism = gds['sample_organism']
            samples = len(get_samples(gds))
            features = gds['variables']
            genes = gds['genes']
            subsets = len(gds['subsets'])

            pubmed = gds.get('pubmed_id', '')
            pubmed_id = pubmed
            if isinstance(pubmed, list) and len(pubmed) > 0:
                pubmed_id = pubmed[0]

            return [
                ' ' if is_cached(gds_id) else '',
                gds_id,
                pubmed_id,
                organism,
                samples,
                features,
                genes,
                subsets,
                title,
            ]

        self.table = np.asarray([__gds_to_row(gds) for gds in info.values()])
        self.show_table()

    def columnCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else self._table.shape[1]

    def clear(self):
        self.beginResetModel()
        self._table = np.array([[]])
        self.resetSorting()
        self._roleData.clear()
        self.endResetModel()

    def data(self, index, role, _str=str,
        _Qt_DisplayRole=Qt.DisplayRole,
        _Qt_EditRole=Qt.EditRole,
        _Qt_FontRole=Qt.FontRole,
        _Qt_ForegroundRole=Qt.ForegroundRole,
        _LinkRolee=LinkRole,
        _recognizedRoles=frozenset([Qt.DisplayRole, Qt.EditRole, Qt.FontRole, Qt.ForegroundRole, LinkRole]),):

        if role not in _recognizedRoles:
            return None

        row, col = index.row(), index.column()
        if not 0 <= row <= self.rowCount():
            return None
        row = self.mapToSourceRows(row)

        try:
            # value = self[row][col]
            value = self._row_instance(row, col)
        except IndexError:
            return

        if role == Qt.DisplayRole:
            return QVariant(str(value))
        elif role == Qt.ToolTipRole:
            return QVariant(str(value))

        if col == self.pubmedid_col:
            if role == _Qt_ForegroundRole:
                return self.color
            elif role == _Qt_FontRole:
                return self.font
            elif role == _LinkRolee:
                return pubmed_url.format(value)

    def get_row_index(self, gds_name):
        # test = self._table[self._table[:, 1] == gds_name, :]
        rows, _ = np.where(np.isin(self._table, gds_name))
        if rows is not None and len(rows) > 0:
            return self.mapFromSourceRows(rows[0])

    def filter_table(self, filter_pattern: str):
        selection = np.full(self.table.shape, True)
        for search_word in filter_pattern.split():
            match_result = np.core.defchararray.find(np.char.lower(self.table), search_word.lower()) >= 0
            selection = selection & match_result
        return selection

    def show_table(self, filter_pattern=''):
        # clear cache if model changes
        self._row_instance.cache_clear()
        self.wrap(self.table[self.filter_table(filter_pattern).any(axis=1), :])


class OWGEODatasets(OWWidget, ConcurrentWidgetMixin):
    name = "GEO Data Sets"
    description = "Access to Gene Expression Omnibus data sets."
    icon = "icons/OWGEODatasets.svg"
    priority = 2

    class Outputs:
        gds_data = Output("Expression Data", Table)

    search_pattern = Setting('')
    auto_commit = Setting(True)
    genes_as_rows = Setting(False)
    mergeSpots = Setting(True)
    selected_gds = Setting(None)
    gdsSelectionStates = Setting({})
    splitter_settings = Setting(
        (
            b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01\xea\x00\x00\x00\xd7\x01\x00\x00\x00\x07\x01\x00\x00\x00\x02',
            b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01\xb5\x00\x00\x02\x10\x01\x00\x00\x00\x07\x01\x00\x00\x00\x01',
        )
    )

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        self.gds_info: Optional[GDSInfo] = GDSInfo()  # TODO: handle possible exceptions
        self.gds_data: Optional[Table] = None

        # Control area
        box = widgetBox(self.controlArea, 'Info', addSpace=True)
        self.infoBox = widgetLabel(box, 'Initializing\n\n')

        box = widgetBox(self.controlArea, 'Output', addSpace=True)
        radioButtonsInBox(box, self, 'genes_as_rows', ['Samples in rows', 'Genes in rows'], callback=self._run)
        separator(box)

        rubber(self.controlArea)
        auto_commit(self.controlArea, self, 'auto_commit', '&Commit', box=False)

        # Main Area

        # Filter widget
        self.filter = lineEdit(self.mainArea, self,
                               'search_pattern',
                               'Filter:',
                               callbackOnType=True,
                               callback=self._apply_filter)
        self.mainArea.layout().addWidget(self.filter)

        splitter_vertical = QSplitter(Qt.Vertical, self.mainArea)

        self.mainArea.layout().addWidget(splitter_vertical)

        # set table view
        self.table_view = QTableView(splitter_vertical)
        self.table_view.setShowGrid(False)
        self.table_view.setSortingEnabled(True)
        self.table_view.sortByColumn(1, Qt.AscendingOrder)
        self.table_view.setAlternatingRowColors(True)
        self.table_view.verticalHeader().setVisible(False)
        self.table_view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table_view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.table_view.viewport().setMouseTracking(True)
        self.table_view.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)

        self.table_model = GEODatasetsModel()
        self.table_model.initialize(self.gds_info)
        self.table_view.setModel(self.table_model)
        self.table_model.show_table()

        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.resizeColumnsToContents()

        v_header = self.table_view.verticalHeader()
        option = self.table_view.viewOptions()
        size = self.table_view.style().sizeFromContents(
            QStyle.CT_ItemViewItem, option, QSize(20, 20), self.table_view
        )

        v_header.setDefaultSectionSize(size.height() + 2)
        v_header.setMinimumSectionSize(5)

        # set item delegates
        self.table_view.setItemDelegateForColumn(
            self.table_model.pubmedid_col, LinkStyledItemDelegate(self.table_view)
        )
        self.table_view.setItemDelegateForColumn(
            self.table_model.gds_id_col, LinkStyledItemDelegate(self.table_view)
        )
        self.table_view.setItemDelegateForColumn(
            self.table_model.indicator_col,
            IndicatorItemDelegate(self.table_view, role=Qt.DisplayRole),
        )

        splitter_horizontal = QSplitter(Qt.Horizontal, splitter_vertical)

        # Description Widget
        box = widgetBox(splitter_horizontal, 'Description')
        self.description_widget = widgetLabel(box, '')
        self.description_widget.setWordWrap(True)
        rubber(box)

        # Sample Annotations Widget
        box = widgetBox(splitter_horizontal, 'Sample Annotations')
        self.annotations_widget = QTreeWidget(box)
        self.annotations_widget.setHeaderLabels(
            ['Type (Sample annotations)', 'Sample count']
        )
        self.annotations_widget.setRootIsDecorated(True)
        box.layout().addWidget(self.annotations_widget)
        self._annotations_updating = False
        self.annotations_widget.itemChanged.connect(self.on_annotation_selection_changed)
        self.splitters = splitter_vertical, splitter_horizontal

        for sp, setting in zip(self.splitters, self.splitter_settings):
            sp.splitterMoved.connect(self._splitter_moved)
            sp.restoreState(setting)

        self.table_view.selectionModel().selectionChanged.connect(self.on_gds_selection_changed)
        self._apply_filter()

        self._run()

    def _splitter_moved(self, *args):
        self.splitter_settings = [bytes(sp.saveState()) for sp in self.splitters]

    def _set_description_widget(self):
        self.description_widget.setText(
            self.selected_gds.get('description', 'Description not available.')
        )

    def _set_annotations_widget(self, gds):
        self._annotations_updating = True
        self.annotations_widget.clear()

        annotations = defaultdict(set)
        subsets_count = {}

        for desc in gds['subsets']:
            annotations[desc['type']].add(desc['description'])
            subsets_count[desc['description']] = str(len(desc['sample_id']))

        for _type, subsets in annotations.items():
            key = (gds["name"], _type)
            parent = QTreeWidgetItem(self.annotations_widget, [_type])
            parent.key = key
            for subset in subsets:
                key = (gds['name'], _type, subset)
                item = QTreeWidgetItem(parent, [subset, subsets_count.get(subset, '')])
                item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
                item.setCheckState(0, self.gdsSelectionStates.get(key, Qt.Checked))
                item.key = key

        self._annotations_updating = False
        self.annotations_widget.expandAll()
        for i in range(self.annotations_widget.columnCount()):
            self.annotations_widget.resizeColumnToContents(i)

    def _set_selection(self):
        if self.selected_gds is not None:
            index = self.table_model.get_row_index(self.selected_gds.get('name'))
            if index is not None:
                self.table_view.selectionModel().blockSignals(True)
                self.table_view.selectRow(index)
                self._handle_selection_changed()
                self.table_view.selectionModel().blockSignals(False)

    def _handle_selection_changed(self):
        if self.table_model.table is not None:
            selection = self.table_view.selectionModel().selectedRows(
                self.table_model.gds_id_col
            )
            selected_gds_name = selection[0].data() if len(selection) > 0 else None

            if selected_gds_name:
                self.selected_gds = self.table_model.info.get(selected_gds_name)
                self._set_annotations_widget(self.selected_gds)
                self._set_description_widget()
            else:
                self.annotations_widget.clear()
                self.description_widget.clear()

            self.update_info()

    def _apply_filter(self):
        if self.table_model.table is not None:
            self.table_model.show_table(filter_pattern=str(self.search_pattern))
            self._set_selection()
            self.update_info()

    def _run(self):
        if self.selected_gds is not None:
            self.gds_data = None
            self.start(run_download_task,
                       self.selected_gds.get('name'),
                       self.get_selected_samples(),
                       self.genes_as_rows)

    def on_gds_selection_changed(self):
        self._handle_selection_changed()
        self._run()

    def on_annotation_selection_changed(self):
        if self._annotations_updating:
            return
        for i in range(self.annotations_widget.topLevelItemCount()):
            item = self.annotations_widget.topLevelItem(i)
            if 'key' in item.__dict__:
                self.gdsSelectionStates[item.key] = item.checkState(0)
            for j in range(item.childCount()):
                child = item.child(j)
                if 'key' in child.__dict__:
                    self.gdsSelectionStates[child.key] = child.checkState(0)

        self._run()

    def update_info(self):
        all_gds = len(self.table_model.info)
        text = "{} datasets\n{} datasets cached\n".format(
            all_gds, len(local_files.listfiles())
        )
        filtered = self.table_view.model().rowCount()
        if all_gds != filtered:
            text += "{} after filtering".format(filtered)
        self.infoBox.setText(text)

    def get_selected_samples(self):
        """
        Return the currently selected sample annotations.

        The return value is a list of selected (sample type, sample value)
        tuples.

        .. note:: if some Sample annotation type has no selected values.
                  this method will return all values for it.

        TODO: this could probably be simplified.

        """

        def childiter(item):
            """ Iterate over the children of an QTreeWidgetItem instance.
            """
            for i in range(item.childCount()):
                yield item.child(i)

        samples = []
        unused_types = []
        used_types = []

        for stype in childiter(self.annotations_widget.invisibleRootItem()):
            selected_values = []
            all_values = []
            for sval in childiter(stype):
                value = (str(stype.text(0)), str(sval.text(0)))
                if self.gdsSelectionStates.get(sval.key, True):
                    selected_values.append(value)
                all_values.append(value)
            if selected_values:
                samples.extend(selected_values)
                used_types.append(str(stype.text(0)))
            else:
                # If no sample of sample type is selected we don't filter on it.
                samples.extend(all_values)
                unused_types.append(str(stype.text(0)))

        _samples = defaultdict(list)
        for sample, sample_type in samples:
            _samples[sample].append(sample_type)
        return _samples

    def commit(self):
        self.Outputs.gds_data.send(self.gds_data)

    def on_done(self, result: Result):
        assert isinstance(result.gds_dataset, Table)
        self.gds_data = result.gds_dataset
        self.commit()

        if self.gds_info:
            self.table_model.initialize(self.gds_info)
            self._apply_filter()

    def on_partial_result(self, result: Any) -> None:
        pass

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def send_report(self):
        self.report_items(
            "GEO Dataset",
            [
                ("ID", self.selected_gds['name']),
                ("Title", self.selected_gds['title']),
                ("Organism", self.selected_gds['sample_organism']),
            ],
        )
        self.report_items(
            "Data",
            [
                ("Samples", self.selected_gds['sample_count']),
                ("Features", self.selected_gds['variables']),
                ("Genes", self.selected_gds['genes']),
            ],
        )
        self.report_name("Sample annotations")
        subsets = defaultdict(list)
        for subset in self.selected_gds['subsets']:
            subsets[subset['type']].append(
                (subset['description'], len(subset['sample_id']))
            )
        self.report_html += "<ul>"
        for _type in subsets:
            self.report_html += "<b>" + _type + ":</b></br>"
            for desc, count in subsets[_type]:
                self.report_html += 9 * "&nbsp" + "<b>{}:</b> {}</br>".format(
                    desc, count
                )
        self.report_html += "</ul>"


if __name__ == "__main__":
    def main_test():
        from AnyQt.QtWidgets import QApplication

        app = QApplication([])
        w = OWGEODatasets()
        w.show()
        w.raise_()
        r = app.exec_()
        w.saveSettings()
        return r


    sys.exit(main_test())



""" Dictybase update """
import bz2

from server_update import *
from server_update.tests.test_Dicty import DictyMutantsTest
from orangecontrib.bioinformatics.dicty import phenotypes
from orangecontrib.bioinformatics.dicty.config import PHENOTYPES_FILENAME, PHENOTYPES_TAGS, DOMAIN, PHENOTYPES_TITLE


""" Update DictyMutants
"""
domain_path = sf_local.localpath(DOMAIN)
temp_path = os.path.join(domain_path, sf_temp)

create_folder(domain_path)
create_folder(temp_path)
localfile = sf_local.localpath(temp_path, PHENOTYPES_FILENAME)

mutants = phenotypes.DictyMutants()
base_mutants = phenotypes.download_mutants()

print("updating dictBase mutants")
with open(localfile, 'wb') as f:
    f.write(base_mutants)
    f.close()

uncompressed = os.stat(os.path.join(domain_path, PHENOTYPES_FILENAME)).st_size  # store uncompressed database size

with bz2.BZ2File(os.path.join(temp_path, PHENOTYPES_FILENAME), mode='w', compresslevel=9) as f_compressed:
    shutil.copyfileobj(open(os.path.join(domain_path, PHENOTYPES_FILENAME), 'rb'), f_compressed)

create_info_file(localfile,
                 title=PHENOTYPES_TITLE,
                 tags=PHENOTYPES_TAGS,
                 source=SOURCE_SERVER,
                 domain=DOMAIN,
                 filename=PHENOTYPES_FILENAME,
                 uncompressed=uncompressed,
                 compression='bz2')

helper = SyncHelper(DOMAIN, DictyMutantsTest)
helper.run_tests()
helper.sync_files()
print("DictyBase mutant updated")

helper.remove_update_folder()



""" Qt component for Gene sets """
import numpy as np

from typing import Union
from collections import defaultdict

from AnyQt.QtWidgets import (
    QTreeView, QTreeWidget, QTreeWidgetItem, QTreeWidgetItemIterator, QWidget,
    QGroupBox
)
from AnyQt.QtCore import Qt

from orangecontrib.bioinformatics.geneset import GeneSet, GeneSets, load_gene_sets, list_all

# TODO: better handle stored selection
# TODO: Don't use hardcoded 'Custom sets', use table name if available


class GeneSetsSelection(QWidget):
    def __init__(self, box, parent, settings_var, **kwargs):
        # type: (Union[QGroupBox, QWidget], QWidget, str) -> None
        super().__init__(**kwargs)

        self.parent = parent
        self.stored_selection = settings_var
        # gene sets object
        self.gs_object = GeneSets()  # type: GeneSets

        self.hierarchy_tree_widget = QTreeWidget(self)
        self.hierarchy_tree_widget.setHeaderHidden(True)
        self.hierarchy_tree_widget.setEditTriggers(QTreeView.NoEditTriggers)
        box.layout().addWidget(self.hierarchy_tree_widget)

        self.custom_set_hier = None
        self.default_selection = [
            ('GO', 'molecular_function'), ('GO', 'biological_process'), ('GO', 'cellular_component')
        ]

    def clear_custom_sets(self):
        # delete any custom sets if they exists
        self.gs_object.delete_sets_by_hierarchy(self.custom_set_hier)

    def add_custom_sets(self, gene_sets_names, gene_names, hierarchy_title=None, select_customs_flag=False):
        # type: (np.ndarray, np.ndarray) -> None

        self.custom_set_hier = hierarchy_title
        self.clear_custom_sets()

        temp_dict = defaultdict(list)
        for set_name, gene_name in zip(gene_sets_names, gene_names):
            temp_dict[set_name].append(gene_name)

        g_sets = []
        for key, value in temp_dict.items():
            g_sets.append(GeneSet(gs_id=key,
                                  hierarchy=self.custom_set_hier,
                                  organism=self.gs_object.common_org(),
                                  name=key,
                                  genes=set(value)))

        self.gs_object.update(g_sets)
        self.update_gs_hierarchy(select_customs_flag=select_customs_flag)

    def load_gene_sets(self, tax_id):
        # type: (str) -> None
        self.gs_object = GeneSets()
        self.clear()

        gene_sets = list_all(organism=tax_id)
        self.set_hierarchy_model(self.hierarchy_tree_widget, self.hierarchy_tree(gene_sets))

        for gene_set in gene_sets:
            g_sets = load_gene_sets(gene_set, tax_id)
            self.gs_object.update([g_set for g_set in g_sets])

        self.set_selected_hierarchies()

    def clear_gene_sets(self):
        self.gs_object = GeneSets()

    def clear(self):
        # reset hierarchy widget state
        self.hierarchy_tree_widget.clear()

    def update_gs_hierarchy(self, select_customs_flag=False):
        self.clear()
        self.set_hierarchy_model(self.hierarchy_tree_widget, self.hierarchy_tree(self.gs_object.hierarchies()))
        if select_customs_flag:
            self.set_custom_sets()
        else:
            self.set_selected_hierarchies()

    def set_hierarchy_model(self, tree_widget, sets):

        def beautify_displayed_text(text):
            if '_' in text:
                return text.replace('_', ' ').title()
            else:
                return text

        # TODO: maybe optimize this code?
        for key, value in sets.items():
            item = QTreeWidgetItem(tree_widget, [beautify_displayed_text(key)])
            item.setFlags(item.flags() & (Qt.ItemIsUserCheckable | Qt.ItemIsSelectable | Qt.ItemIsEnabled))
            item.setExpanded(True)
            item.hierarchy = key

            if value:
                item.setFlags(item.flags() | Qt.ItemIsTristate)
                self.set_hierarchy_model(item, value)
            else:
                if item.parent():
                    item.hierarchy = (item.parent().hierarchy, key)

            if not item.childCount() and not item.parent():
                item.hierarchy = (key, )

    def get_hierarchies(self, **kwargs):
        """ return selected hierarchy
        """
        only_selected = kwargs.get('only_selected', None)

        sets_to_display = list()

        if only_selected:
            iterator = QTreeWidgetItemIterator(self.hierarchy_tree_widget, QTreeWidgetItemIterator.Checked)
        else:
            iterator = QTreeWidgetItemIterator(self.hierarchy_tree_widget)

        while iterator.value():
            # note: if hierarchy value is not a tuple, then this is just top level qTreeWidgetItem that
            #       holds subcategories. We don't want to display all sets from category
            if type(iterator.value().hierarchy) is not str:

                if not only_selected:
                    sets_to_display.append(iterator.value().hierarchy)
                else:
                    if not iterator.value().isDisabled():
                        sets_to_display.append(iterator.value().hierarchy)

            iterator += 1

        return sets_to_display

    def set_selected_hierarchies(self):
        iterator = QTreeWidgetItemIterator(self.hierarchy_tree_widget, QTreeWidgetItemIterator.All)
        defaults = []

        while iterator.value():

            # note: if hierarchy value is not a tuple, then this is just top level qTreeWidgetItem that
            #       holds subcategories. We don't want to display all sets from category
            if type(iterator.value().hierarchy) is not str:
                if iterator.value().hierarchy in self.parent.__getattribute__(self.stored_selection):
                    iterator.value().setCheckState(0, Qt.Checked)
                else:
                    iterator.value().setCheckState(0, Qt.Unchecked)

            # if no items are checked, set defaults
            if iterator.value().hierarchy in self.default_selection:
                defaults.append(iterator.value())

            iterator += 1

        if len(self.get_hierarchies(only_selected=True)) == 0:
            [item.setCheckState(0, Qt.Checked) for item in defaults]

    def set_custom_sets(self):
        iterator = QTreeWidgetItemIterator(self.hierarchy_tree_widget, QTreeWidgetItemIterator.All)

        while iterator.value():

            # note: if hierarchy value is not a tuple, then this is just top level qTreeWidgetItem that
            #       holds subcategories. We don't want to display all sets from category
            if type(iterator.value().hierarchy) is not str:
                if iterator.value().hierarchy == self.custom_set_hier:
                    iterator.value().setCheckState(0, Qt.Checked)
                else:
                    iterator.value().setCheckState(0, Qt.Unchecked)

            iterator += 1

    @staticmethod
    def hierarchy_tree(gene_sets):
        def tree():
            return defaultdict(tree)

        collection = tree()

        def collect(col, set_hierarchy):
            if set_hierarchy:
                collect(col[set_hierarchy[0]], set_hierarchy[1:])

        for hierarchy in gene_sets:
            collect(collection, hierarchy)

        return collection



import unittest

import numpy as np
from Orange.data import Table, Domain, StringVariable, ContinuousVariable

from orangecontrib.bioinformatics.annotation.annotate_samples import \
    AnnotateSamples, SCORING_EXP_RATIO, SCORING_MARKERS_SUM, SCORING_LOG_FDR, \
    SCORING_LOG_PVALUE
from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID


class TestAnnotateSamples(unittest.TestCase):

    def setUp(self):
        m_domain = Domain(
            [], None, [
                StringVariable("Cell Type"), StringVariable("Entrez ID")])
        m_data = [["Type 1", "111"], ["Type 1", "112"], ["Type 1", "113"],
                  ["Type 1", "114"],
                  ["Type 2", "211"], ["Type 2", "212"], ["Type 2", "213"],
                  ["Type 2", "214"],
                  ]
        self.markers = Table(m_domain, np.empty((len(m_data), 0)), None, m_data)

        genes = ["111", "112", "113", "114", "211", "212", "213", "214"]
        self.domain = Domain([ContinuousVariable(
            str(g)) for g in genes])
        for v, g in zip(self.domain.attributes, genes):
            v.attributes = {"Entrez ID": g}
        self.data = Table(self.domain, np.array(
            [[1, 1, 1, 1.1, 0, 0, 0, 0],
             [1, .8, .9, 1, 0, 0, 0, 0],
             [.7, 1.1, 1, 1.2, 0, 0, 0, 0],
             [.8, .7, 1.1, 1, 0, .1, 0, 0],
             [0, 0, 0, 0, 1.05, 1.05, 1.1, 1],
             [0, 0, 0, 0, 1.1, 1.0, 1.05, 1.1],
             [0, 0, 0, 0, 1.05, .9, 1.1, 1.1],
             [0, 0, 0, 0, .9, .9, 1.2, 1]
             ]))
        self.data.attributes[TAX_ID] = "9606"  # id for a human

        self.annotator = AnnotateSamples()

    def test_artificial_data(self):
        annotations = self.annotator.annotate_samples(self.data, self.markers)

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 2)  # two types in the data
        self.assertGreater(np.nansum(annotations.X), 0)
        self.assertLessEqual(np.nanmax(annotations.X), 1)
        self.assertGreaterEqual(np.nanmin(annotations.X), 0)

    def test_remove_empty_column(self):
        """
        Type 3 column must be removed here
        """
        m_domain = Domain(
            [], None, [
                StringVariable("Cell Type"), StringVariable("Entrez ID")])
        m_data = [["Type 1", "111"], ["Type 1", "112"], ["Type 1", "113"],
                  ["Type 1", "114"],
                  ["Type 2", "211"], ["Type 2", "212"], ["Type 2", "213"],
                  ["Type 2", "214"],
                  ["Type 3", "311"], ["Type 3", "312"], ["Type 3", "313"]]
        markers = Table(m_domain, np.empty((len(m_data), 0)), None, m_data)

        annotations = self.annotator.annotate_samples(self.data, markers)

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 2)  # two types in the data
        self.assertGreater(np.nansum(annotations.X), 0)
        self.assertLessEqual(np.nanmax(annotations.X), 1)
        self.assertGreaterEqual(np.nanmin(annotations.X), 0)

        annotations = self.annotator.annotate_samples(
            self.data, markers, return_nonzero_annotations=False)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 3)  # two types in the data
        self.assertGreater(np.nansum(annotations.X), 0)
        self.assertLessEqual(np.nanmax(annotations.X), 1)
        self.assertGreaterEqual(np.nanmin(annotations.X), 0)

    def test_sf(self):
        """
        Test annotations with hypergeom.sf
        """
        annotator = AnnotateSamples()
        annotations = annotator.annotate_samples(
            self.data, self.markers, p_value_fun="TEST_HYPERGEOMETRIC")

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 2)  # two types in the data
        self.assertGreater(np.nansum(annotations.X), 0)
        self.assertLessEqual(np.nanmax(annotations.X), 1)
        self.assertGreaterEqual(np.nanmin(annotations.X), 0)

    def test_two_example(self):
        self.data.X = self.data.X[:2]

        annotations = self.annotator.annotate_samples(self.data, self.markers)

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))

    def test_markers_without_entrez_id(self):
        self.markers[1, "Entrez ID"] = "?"
        annotations = self.annotator.annotate_samples(
            self.data, self.markers, return_nonzero_annotations=False)

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 2)  # two types in the data
        self.assertGreater(np.nansum(annotations.X), 0)
        self.assertLessEqual(np.nanmax(annotations.X), 1)
        self.assertGreaterEqual(np.nanmin(annotations.X), 0)

    def test_select_attributes(self):
        z = self.annotator.mann_whitney_test(self.data)

        self.assertEqual(z.X.shape, self.data.X.shape)
        self.assertGreaterEqual(z.X[0, 0], 1)
        self.assertGreaterEqual(z.X[0, 1], 1)
        self.assertGreaterEqual(z.X[0, 3], 1)

    def test_assign_annotations(self):
        z = np.array([
            [1.1, 1.1, 1.1, 1.1, 0, 0, 0, 0],
            [1.1, 1.1, 0, 0, 1.1, 0, 0, 0],
            [1.1, 0, 0, 0, 1.1, 1.1, 0, 0],
            [0, 0, 0, 0, 1.1, 1.1, 1.1, 1.1]
        ])
        z_table = Table(self.domain, z)
        attrs = [{"111", "112", "113", "114"}, {"111", "112", "211"},
                 {"211", "212", "111"}, {"211", "212", "213", "214"}]
        exp_ann = np.array([
            [1, 0],
            [1/2, 1/4],
            [1/4, 1/2],
            [0, 1]])
        annotations, fdrs = self.annotator.assign_annotations(
            z_table, self.markers, self.data)

        self.assertEqual(len(attrs), len(annotations))
        self.assertEqual(len(attrs), len(fdrs))
        self.assertEqual(2, annotations.X.shape[1])  # only two types in markers
        self.assertEqual(2, fdrs.X.shape[1])
        np.testing.assert_array_almost_equal(exp_ann, annotations)

        exp_fdrs_smaller = np.array([
            [0.05, 2],
            [2, 2],
            [2, 2],
            [2, 0.05]])

        np.testing.assert_array_less(fdrs, exp_fdrs_smaller)

    def test_scoring(self):
        # scoring SCORING_EXP_RATIO
        annotations = self.annotator.annotate_samples(
            self.data, self.markers, scoring=SCORING_EXP_RATIO)

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 2)  # two types in the data
        self.assertGreater(np.nansum(annotations.X), 0)
        self.assertLessEqual(np.nanmax(annotations.X), 1)
        self.assertGreaterEqual(np.nanmin(annotations.X), 0)

        # scoring SCORING_MARKERS_SUM
        annotations = self.annotator.annotate_samples(
            self.data, self.markers, scoring=SCORING_MARKERS_SUM)

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 2)  # two types in the data

        # based on provided data it should match
        # the third row is skipped, since it is special
        self.assertEqual(annotations[0, 0].value, self.data.X[0].sum())
        self.assertEqual(annotations[5, 1].value, self.data.X[5].sum())

        # scoring SCORING_LOG_FDR
        annotations = self.annotator.annotate_samples(
            self.data, self.markers, scoring=SCORING_LOG_FDR)

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 2)  # two types in the data

        # scoring SCORING_LOG_PVALUE
        annotations = self.annotator.annotate_samples(
            self.data, self.markers, scoring=SCORING_LOG_PVALUE)

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 2)  # two types in the data

    def test_log_cpm(self):
        norm_data = self.annotator.log_cpm(self.data)
        self.assertTupleEqual(self.data.X.shape, norm_data.X.shape)

    def test_entrez_id_not_string(self):
        """
        It seems that some datasets (e.g. AML dataset) have Entrez ID as int
        although they should be strings. Here we add the test for those cases.
        """
        # change Entrez IDs to int
        for i, att in enumerate(self.data.domain.attributes):
            att.attributes["Entrez ID"] = int(att.attributes["Entrez ID"])

        annotations = self.annotator.annotate_samples(self.data,
                                                      self.markers)

        self.assertEqual(type(annotations), Table)
        self.assertEqual(len(annotations), len(self.data))
        self.assertEqual(len(annotations[0]), 2)  # two types in the data
        self.assertGreater(np.nansum(annotations.X), 0)
        self.assertLessEqual(np.nanmax(annotations.X), 1)
        self.assertGreaterEqual(np.nanmin(annotations.X), 0)




""" List Completer """
import re


from AnyQt.QtWidgets import QCompleter
from AnyQt.QtCore import QStringListModel, QObject, Qt


class TokenListCompleter(QCompleter):
    """
    A completer allowing completion of multiple words (tokens)

    Example
    -------
    completer = TokenListCompleter()
    completer.setTokenList(["foo", "bar", "baz"])
    completer.setCompletionPrefix("foo b")
    for i in range(completer.completionCount()):
         completer.setCurrentRow(i)
         print(completer.currentCompletion())

    """
    def __init__(self, *args, **kwargs):
        separator = kwargs.pop("separator", " ")
        super().__init__(*args, **kwargs)
        self.__tokenList = []
        self.__completerModel = None
        self.__separator = separator
        # The current 'known' completion prefix (tracked in splitPath)
        self.__currentKnownPrefix = ""
        self.setModelSorting(QCompleter.CaseSensitivelySortedModel)
        super().setModel(QStringListModel(self))

    def setTokenList(self, tokenList):
        """
        Set the token word completion list

        Equivalent to
        `completer.setModel(QStringListModel(tokenList, completer))`

        Note that this is the preferred method for setting the completion
        model.

        Parameters
        ----------
        tokenList : List[str]
        """
        tokenList = list(sorted(set(tokenList)))
        self.setModel(QStringListModel(tokenList, self))

    def tokenList(self):
        """
        Return the current token word completion list.

        Return
        ------
        tokens : List[str]
        """
        return list(self.__tokenList)

    def setModel(self, model):
        """
        Reimplemented.

        Parameters
        ----------
        model : QAbstractItemModel
        """
        if model is self.__completerModel:
            return

        if self.__completerModel is not None:
            self.__completerModel.dataChanged.disconnect(
                self.__initDynamicModel)
            self.__completerModel.rowsInserted.disconnect(
                self.__initDynamicModel)
            self.__completerModel.rowsRemoved.disconnect(
                self.__initDynamicModel)

            if QObject.parent(self.__completerModel) is self:
                self.__completerModel.deleteLater()
            self.__completerModel = None

        self.__completerModel = model

        if self.__completerModel is not None:
            self.__completerModel.dataChanged.connect(
                self.__initDynamicModel)
            self.__completerModel.rowsInserted.connect(
                self.__initDynamicModel)
            self.__completerModel.rowsRemoved.connect(
                self.__initDynamicModel)

        self.__initDynamicModel()

    def model(self):
        return self.__completerModel

    def setSeparator(self, separator):
        if self.__separator != separator:
            self.__separator = separator

    def separator(self):
        return self.__separator

    # QCompleter::setCompleterPrefix is not a virtual method, and is called
    # directly by the QLineEdit to set the prefix. There does not appear to
    # be any other direct way that a QCompleter subclass could be notified
    # of the completion prefix change.
    # However in QCompleter::setCompleterPrefix
    # the prefix is immediately split by splitPath which is virtual.
    # So we use this to track changes in the current prefix.
    def splitPath(self, prefix):
        """reimplemented."""
        items = super().splitPath(prefix)
        if self.__currentKnownPrefix != self.completionPrefix():
            self.__currentKnownPrefix = self.completionPrefix()
            self.__updateEffectiveCompleterModel()
        return items

    def __updateEffectiveCompleterModel(self):
        """
        Prefix all items in the effective completer model with the current
        prefix to enable the completion of multiple keywords.
        """
        prefix = self.completionPrefix()
        model = super().model()
        assert isinstance(model, QStringListModel)
        separator = self.__separator
        if not prefix.rstrip().endswith(separator) and separator in prefix:
            # We are in the middle of a second (or later) word completion
            prefix, rest = prefix.rsplit(separator, 1)
            # preserve any whitespace after the seperator
            match = re.match(r"^(?P<whitespace>\s*)(?P<rest>.*)", rest)
            if match:
                whitespace, rest = match.groups()
            else:
                whitespace = ""
            items = [prefix + separator + whitespace + item
                     for item in self.__tokenList]
        else:
            # Either completing the the first word or immediately after
            # the separator. In both cases the original tokens list is
            # the active completion list
            items = self.__tokenList
        itemset = set(items)
        current = model.stringList()

        if itemset != set(current):
            model.setStringList(list(sorted(itemset)))

    def __initDynamicModel(self):
        """
        [Re]Initialize the effective completion model from the client
        supplied model
        """
        model = self.__completerModel
        if model is not None:
            if isinstance(model, QStringListModel):
                tokens = model.stringList()
            else:
                tokens = [model.data(model.index(row, 0), Qt.DisplayRole)
                          for row in range(model.rowCount())]
                tokens = [str(token) for token in filter(None, tokens)]
        else:
            tokens = []

        tokenList = list(sorted(set(tokens)))
        if tokenList != self.__tokenList:
            self.__tokenList = tokenList
            self.__updateEffectiveCompleterModel()



""" Documentation script """
from Orange.classification import LogisticRegressionLearner
from Orange.evaluation.testing import CrossValidation
from Orange.evaluation.scoring import AUC

from orangecontrib.bioinformatics.geo.dataset import GDS

gds = GDS("GDS2960")
data = gds.get_data(sample_type="disease state", transpose=True, report_genes=True)
print("Samples: %d, Genes: %d" % (len(data), len(data.domain.attributes)))

learners = [LogisticRegressionLearner()]
results = CrossValidation(data, learners, k=10)

print("AUC = %.3f" % AUC(results)[0])



# pylint: disable=too-many-ancestors
from enum import IntEnum
from typing import Tuple, Optional
from types import SimpleNamespace as namespace
from itertools import chain
import numpy as np

from AnyQt.QtCore import Qt, QRectF, QObject
from AnyQt.QtGui import QColor

import pyqtgraph as pg

from Orange.data import Table, Domain
from Orange.data.filter import FilterString, Values
from Orange.projection import Projector, PCA
from Orange.widgets import gui, report
from Orange.widgets.settings import Setting, SettingProvider
from Orange.widgets.unsupervised.owtsne import pca_preprocessing, \
    prepare_tsne_obj
from Orange.widgets.utils.colorpalette import ColorPaletteGenerator
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.visualize.owscatterplotgraph import OWScatterPlotBase
from Orange.widgets.visualize.utils.widget import OWDataProjectionWidget
from Orange.widgets.widget import Input, Msg

from orangecontrib.bioinformatics.annotation.annotate_projection import \
    annotate_projection
from orangecontrib.bioinformatics.annotation.annotate_samples import \
    AnnotateSamples, SCORING_EXP_RATIO, SCORING_MARKERS_SUM, \
    SCORING_LOG_FDR, PFUN_BINOMIAL, PFUN_HYPERGEOMETRIC
from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID


class Result(namespace):
    scores = None  # type: Optional[Table]
    embedding = None  # type: Optional[Table]
    clusters = None  # type: Clusters


class Runner:
    @staticmethod
    def compute_scores(data: Table, genes: Table, p_threshold: float,
                       p_value_fun: str, scoring: str, start: float,
                       end: float, result: Result, state: TaskState):
        if not data or not genes:
            result.scores.z_vals = None
            result.scores.annotations = None
            result.scores.p_vals = None
            result.scores.table = None
        else:
            state.set_status("Computing scores...")
            weights = np.array([15, 75, 10]) * (end - start) / 100

            if not result.scores.z_vals:
                result.scores.z_vals = AnnotateSamples.mann_whitney_test(data)
                state.set_partial_result(("scores", result))
            state.set_progress_value(weights[0])
            if state.is_interruption_requested():
                return

            if not result.scores.annotations or not result.scores.p_vals:
                annot, p_vals = AnnotateSamples.assign_annotations(
                    result.scores.z_vals, genes, data,
                    p_value_fun=p_value_fun, scoring=scoring)
                result.scores.annotations = annot
                result.scores.p_vals = p_vals
                state.set_partial_result(("scores", result))
            state.set_progress_value(weights[1])
            if state.is_interruption_requested():
                return

            result.scores.table = AnnotateSamples.filter_annotations(
                result.scores.annotations, result.scores.p_vals,
                p_threshold=p_threshold)

        state.set_partial_result(("scores", result))

    @staticmethod
    def compute_tsne_embedding(data: Table, start: float, end: float,
                               state: TaskState):
        count, weights = (0, np.array([7, 1, 1, 5, 1, 23, 61, 1]) *
                          (end - start) / 100)
        tsne = prepare_tsne_obj(data, 30, False, 1)

        def set_progress():
            nonlocal start
            nonlocal count
            start += weights[count]
            count += 1
            state.set_progress_value(start)
            return 0 if state.is_interruption_requested() else 1

        pca_data = pca_preprocessing(data, 20, True)
        if not set_progress():
            return None

        pp_data = tsne.preprocess(pca_data)
        if not set_progress():
            return None

        init = tsne.compute_initialization(pp_data.X)
        if not set_progress():
            return None

        affinities = tsne.compute_affinities(pp_data.X)
        if not set_progress():
            return None

        tsne_embedding = tsne.prepare_embedding(affinities, init)
        if not set_progress():
            return None

        def tsne_optimization(n_iter, **kwargs):
            nonlocal tsne_embedding
            step_size = 25
            n_iter_done = 0
            while n_iter_done < n_iter:
                n_iter_done += step_size
                tsne_embedding = tsne_embedding.optimize(
                    n_iter=step_size, inplace=True, **kwargs)
                step = n_iter_done / n_iter * weights[count]
                state.set_progress_value(start + step)
                if state.is_interruption_requested():
                    return None

        kwargs = {"exaggeration": tsne.early_exaggeration, "momentum": 0.5}
        tsne_optimization(tsne.early_exaggeration_iter, **kwargs)
        if not set_progress():
            return None

        kwargs = {"exaggeration": tsne.exaggeration, "momentum": 0.8}
        tsne_optimization(tsne.n_iter, **kwargs)
        if not set_progress():
            return None

        tsne_embedding = tsne.convert_embedding_to_model(
            pp_data, tsne_embedding)
        set_progress()
        return tsne_embedding.embedding

    @staticmethod
    def compute_embedding(data: Table, projector: Optional[Projector],
                          start: float, end: float,
                          result: Result, state: TaskState):
        if not data:
            result.embedding = None
        else:
            state.set_status("Computing 2D projection...")
            if projector is None:
                # Use t-SNE by default
                embedding = Runner.compute_tsne_embedding(
                    data, start, end, state)
            elif isinstance(projector, PCA):
                embedding = projector(data)(data)[:, :2]
            else:
                raise NotImplementedError
            result.embedding = embedding
        state.set_partial_result(("embedding", result))

    @staticmethod
    def compute_clusters(result: Result, state: TaskState):
        if not result.scores.table or not result.embedding:
            result.clusters.table = None
            result.clusters.groups = None
        else:
            state.set_status("Finding clusters...")
            kwargs = {}
            if result.clusters.epsilon is not None:
                kwargs["eps"] = result.clusters.epsilon
            clusters = annotate_projection(
                result.scores.table, result.embedding, **kwargs)
            result.clusters.table = clusters[0]
            result.clusters.groups = clusters[1]
            result.clusters.epsilon = clusters[2]
        state.set_partial_result(("clusters", result))

    @classmethod
    def run(cls, data: Table, genes: Table, projector: Optional[Projector],
            p_threshold: float, p_value_fun: str, scoring: str,
            result: Result, state: TaskState):

        start, step, weights = 0, 0, np.array([25, 70, 5])

        def set_progress():
            nonlocal start
            nonlocal step
            start = int(start + weights[step])
            step += 1
            state.set_progress_value(start)
            return 0 if state.is_interruption_requested() else 1

        if not result.scores.table:
            end = start + weights[step]
            cls.compute_scores(data, genes, p_threshold, p_value_fun,
                               scoring, start, end, result, state)
        if not set_progress():
            return result

        if not result.embedding:
            end = start + weights[step]
            cls.compute_embedding(data, projector, start, end, result, state)
        if not set_progress():
            return result

        if not result.clusters.table:
            cls.compute_clusters(result, state)
        set_progress()
        return result


class CenteredTextItem(pg.TextItem):
    def __init__(self, view_box, x, y, text, tooltip, rgb):
        bg_color = QColor(*rgb)  # TODO - remove rgb, or use it
        bg_color = QColor(Qt.white)
        bg_color.setAlpha(200)
        color = QColor(Qt.black)
        super().__init__(text, pg.mkColor(color), fill=pg.mkBrush(bg_color))
        self._x = x
        self._y = y
        self._view_box = view_box
        self._view_box.sigStateChanged.connect(self.center)
        self.textItem.setToolTip(tooltip)
        self.setPos(x, y)
        self.center()

    def center(self):
        br = self.boundingRect()
        dx = br.width() / 2 * self._view_box.viewPixelSize()[0]
        dy = br.height() / 2 * self._view_box.viewPixelSize()[1]
        self.setPos(self._x - dx, self._y + dy)


class EventDelegate(QObject):
    def eventFilter(self, *_):
        return False


class OWAnnotateProjectionGraph(OWScatterPlotBase):
    show_cluster_hull = Setting(True)
    n_cluster_labels = Setting(1)

    def __init__(self, scatter_widget, parent):
        super().__init__(scatter_widget, parent)
        self.cluster_hulls_items = []
        self.cluster_labels_items = []
        self._tooltip_delegate = EventDelegate()  # remove points tooltip

    def clear(self):
        super().clear()
        self.cluster_hulls_items.clear()
        self.cluster_labels_items.clear()

    def reset_view(self):
        x, y = self.get_coordinates()
        hulls = self.master.get_cluster_hulls()
        if not hulls or x is None or y is None:
            return
        x = np.hstack([hull[:, 0] for hull, _ in hulls] + [x])
        min_x, max_x = np.min(x), np.max(x)
        y = np.hstack([hull[:, 1] for hull, _ in hulls] + [y])
        min_y, max_y = np.min(y), np.max(y)
        rect = QRectF(min_x, min_y, max_x - min_x or 1, max_y - min_y or 1)
        self.view_box.setRange(rect, padding=0.025)

    def update_coordinates(self):
        super().update_coordinates()
        if self.scatterplot_item is not None:
            self.update_cluster_hull()
            self.update_cluster_labels()
            self.view_box.setAspectLocked(True, 1)

    def update_cluster_hull(self):
        for item in self.cluster_hulls_items:
            self.plot_widget.removeItem(item)
        if not self.show_cluster_hull:
            return
        hulls = self.master.get_cluster_hulls()
        if hulls is None:
            return
        for hull, color in hulls:
            pen = pg.mkPen(color=QColor(*color), style=Qt.DashLine, width=3)
            item = pg.PlotCurveItem(
                x=hull[:, 0], y=hull[:, 1], pen=pen, antialias=True)
            self.plot_widget.addItem(item)
            self.cluster_hulls_items.append(item)

    def update_cluster_labels(self):
        for item in self.cluster_labels_items:
            self.plot_widget.removeItem(item)
        if not self.n_cluster_labels:
            return
        labels = self.master.get_cluster_labels()
        if labels is None:
            return
        for label_per, (xp, yp), color in labels:
            text = "\n".join([l for l, _ in label_per[:self.n_cluster_labels]])
            ttip = "\n".join([f"{round(p * 100)}%  {l}" for l, p in label_per])
            item = CenteredTextItem(self.view_box, xp, yp, text, ttip, color)
            self.plot_widget.addItem(item)
            self.cluster_labels_items.append(item)


class ScoringMethod(IntEnum):
    ExpRatio, MarkersSum, LogFdr = range(3)

    @staticmethod
    def values():
        return [SCORING_LOG_FDR, SCORING_MARKERS_SUM, SCORING_EXP_RATIO]

    @staticmethod
    def items():
        return ["-Log(FDR)", "Marker expression", "Marker expression (%)"]


class StatisticalTest(IntEnum):
    Binomial, Hypergeometric = range(2)

    @staticmethod
    def values():
        return [PFUN_BINOMIAL, PFUN_HYPERGEOMETRIC]

    @staticmethod
    def items():
        return ["Binomial", "Hypergeometric"]


class OWAnnotateProjection(OWDataProjectionWidget, ConcurrentWidgetMixin):
    name = "Annotator"
    description = "Annotates projection clusters."
    icon = "icons/OWAnnotateProjection.svg"
    priority = 3050
    keywords = ["annotate", "projection", "tsne", "pca"]

    GRAPH_CLASS = OWAnnotateProjectionGraph
    graph = SettingProvider(OWAnnotateProjectionGraph)
    left_side_scrolling = True  # remove this line when https://github.com/biolab/orange3/pull/3863 is merged

    scoring_method = Setting(ScoringMethod.ExpRatio)
    statistical_test = Setting(StatisticalTest.Binomial)
    p_threshold = Setting(0.05)
    use_user_epsilon = Setting(False)
    epsilon = Setting(0)
    color_by_cluster = Setting(False)

    class Scores(namespace):
        z_vals = None  # type: Optional[Table]
        annotations = None  # type: Optional[Table]
        p_vals = None  # type: Optional[Table]
        table = None  # type: Optional[Table]

    class Clusters(namespace):
        table = None  # type: Optional[Table]
        groups = None  # type: Optional[Dict[str, Tuple]]
        epsilon = None  # type: Optional[float]

    class Inputs(OWDataProjectionWidget.Inputs):
        genes = Input("Genes", Table)
        projector = Input("Projector", Projector)

    class Information(OWDataProjectionWidget.Information):
        modified = Msg("The parameter settings have been changed. Press "
                       "\"Start\" to rerun with the new settings.")

    class Warning(OWDataProjectionWidget.Warning):
        no_genes = Msg("Missing genes table on input.")

    class Error(OWDataProjectionWidget.Error):
        proj_error = Msg("An error occurred while annotating data.\n{}")
        missing_tax_id = Msg(f"'{TAX_ID}' is missing in data table. "
                             f"Try using 'Genes' widget.")
        missing_entrez_id = Msg("'Entred ID' is missing in genes table.")
        missing_cell_type = Msg("'Cell Type' is missing in genes table.")

    def __init__(self):
        OWDataProjectionWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)
        # inputs
        self.projector = None  # type: Optional[Projector]
        self.genes = None  # type: Optional[Table]
        # annotations
        self.scores = None  # type: Scores
        self.embedding = None  # type: Optional[Table]
        self.clusters = None  # type: Clusters
        self.__invalidate_scores()
        self.__invalidate_clusters()

    # GUI
    def _add_controls(self):
        self.__add_annotation_controls()
        super()._add_controls()
        self.gui.add_control(
            self._effects_box, gui.hSlider, "Cluster labels:",
            master=self.graph, value="n_cluster_labels", minValue=0,
            maxValue=3, step=1,
            createLabel=False, callback=self.graph.update_cluster_labels
        )
        self._plot_box.children()[1].hide()  # Hide 'Show color regions'
        gui.checkBox(
            self._plot_box, self.graph, "show_cluster_hull",
            "Show cluster hull", callback=self.graph.update_cluster_hull)
        gui.checkBox(
            self._plot_box, self, "color_by_cluster",
            "Color points by cluster",
            callback=self.__color_by_cluster_changed)

    def __add_annotation_controls(self):
        box = gui.vBox(self.controlArea, True)
        gui.comboBox(
            box, self, "scoring_method", label="Scoring method:",
            items=ScoringMethod.items(), orientation=Qt.Horizontal,
            contentsLength=13, labelWidth=100,
            callback=self.__scoring_combo_changed)
        gui.comboBox(
            box, self, "statistical_test", label="Statistical test:",
            items=StatisticalTest.items(), orientation=Qt.Horizontal,
            labelWidth=100, callback=self.__scoring_combo_changed)
        gui.doubleSpin(
            box, self, "p_threshold", 0, 1, 0.01, label="FDR threshold:",
            callback=self.__p_threshold_changed)
        hbox = gui.hBox(box)
        gui.checkBox(
            hbox, self, "use_user_epsilon", "ε for DBSCAN:",
            callback=self.__epsilon_check_changed)
        self.epsilon_spin = gui.doubleSpin(
            hbox, self, "epsilon", 0, 10, 0.1, callback=self.__epsilon_changed)
        self.run_button = gui.button(box, self, "Start", self._toggle_run)

    def __color_by_cluster_changed(self):
        self.controls.attr_color.setEnabled(not self.color_by_cluster)
        self.graph.update_colors()

    def __scoring_combo_changed(self):
        self.__invalidate_scores_annotations()
        self.__parameter_changed()

    def __p_threshold_changed(self):
        self.__invalidate_scores_table()
        self.__parameter_changed()

    def __epsilon_changed(self):
        self.__parameter_changed()

    def __parameter_changed(self):
        self.__invalidate_clusters()
        self.Information.modified()
        self.Error.proj_error.clear()

    def __epsilon_check_changed(self):
        self.enable_epsilon_spin()
        if not self.use_user_epsilon:
            self.__epsilon_changed()

    def __invalidate_scores(self):
        self.scores = self.Scores(z_vals=None, annotations=None,
                                  p_vals=None, scores=None)

    def __invalidate_scores_table(self):
        self.scores.table = None

    def __invalidate_scores_annotations(self):
        self.scores.annotations = None
        self.scores.p_vals = None
        self.__invalidate_scores_table()

    def __invalidate_clusters(self):
        epsilon = self.epsilon if self.use_user_epsilon else None
        self.clusters = self.Clusters(table=None, groups=None, epsilon=epsilon)

    def _toggle_run(self):
        if self.task is not None:
            self.cancel()
            self.run_button.setText("Resume")
            self.commit()
        else:
            self._run()

    def _run(self):
        self.Information.modified.clear()
        if not self.data:
            return

        # Remove cluster hulls and labels
        self.graph.update_cluster_hull()
        self.graph.update_cluster_labels()

        self.run_button.setText("Stop")
        result = Result(scores=self.scores,
                        embedding=self.embedding,
                        clusters=self.clusters)
        self.start(Runner.run, self.data, self.genes, self.projector,
                   self.p_threshold,
                   StatisticalTest.values()[self.statistical_test],
                   ScoringMethod.values()[self.scoring_method], result)

    def on_partial_result(self, value: Tuple[str, Result]):
        which, result = value
        if which == "scores":
            self.scores = result.scores
        elif which == "embedding":
            self.embedding = result.embedding
            self.setup_plot()
        elif which == "clusters":
            self.clusters = result.clusters
            if result.clusters.epsilon is not None:
                self.epsilon = result.clusters.epsilon
            self.graph.update_cluster_hull()
            self.graph.update_cluster_labels()
            self.graph.update_colors()
            self.graph.reset_view()

    def on_done(self, result: Result):
        self.scores = result.scores
        self.embedding = result.embedding
        self.clusters = result.clusters
        if result.clusters.epsilon is not None:
            self.epsilon = result.clusters.epsilon
        self.run_button.setText("Start")
        self.commit()

    def on_exception(self, ex: Exception):
        self.Error.proj_error(ex)
        self.run_button.setText("Start")

    def check_data(self):
        super().check_data()
        if self.data:
            if TAX_ID not in self.data.attributes:
                self.Error.missing_tax_id()
                self.data = None
        if self.data and not self.genes:
            self.Warning.no_genes()

    def check_genes(self):
        if self.genes:
            metas = [m.name for m in self.genes.domain.metas]
            if "Entrez ID" not in metas:
                self.Error.missing_entrez_id()
                self.genes = None
            elif "Cell Type" not in metas:
                self.Error.missing_cell_type()
                self.genes = None

    def enable_controls(self):
        super().enable_controls()
        self.enable_epsilon_spin()
        self.controls.attr_color.setEnabled(not self.color_by_cluster)

    def enable_epsilon_spin(self):
        self.epsilon_spin.setEnabled(self.use_user_epsilon)

    @Inputs.genes
    def set_genes(self, genes: Optional[Table]):
        self.genes = genes
        self.__invalidate_scores_annotations()
        self.__invalidate_clusters()
        self.check_data()
        self.check_genes()

    @Inputs.projector
    def set_projector(self, projector: Optional[Projector]):
        self.cancel()
        self.embedding = None
        self._invalidated = True
        self.__invalidate_clusters()
        self.projector = projector

    def handleNewSignals(self):
        self._run()
        super().handleNewSignals()

    def get_coordinates_data(self):
        coordinates = super().get_coordinates_data()
        if coordinates[0] is not None and len(coordinates) == 1:
            return np.vstack((coordinates, np.zeros_like(coordinates)))
        return coordinates

    def get_embedding(self):
        if not self.embedding:
            self.valid_data = None
            return None

        self.valid_data = np.all(np.isfinite(self.embedding.X), 1)
        return self.embedding.X

    def get_color_data(self):
        if not self.color_by_cluster or not self.clusters.table:
            return super().get_color_data()

        attr = self.clusters.table.domain["Clusters"]
        all_data = self.clusters.table.get_column_view(attr)[0]
        if all_data.dtype == object and attr.is_primitive():
            all_data = all_data.astype(float)
        if self.valid_data is not None:
            all_data = all_data[self.valid_data]
        return all_data

    def get_color_labels(self):
        if not self.color_by_cluster or not self.clusters.table:
            return super().get_color_labels()
        return self.clusters.table.domain["Clusters"].values

    def is_continuous_color(self):
        if not self.color_by_cluster or not self.clusters.table:
            return super().is_continuous_color()
        return False

    def get_palette(self):
        if not self.color_by_cluster or not self.clusters.table:
            return super().get_palette()

        colors = self.clusters.table.domain["Clusters"].colors
        return ColorPaletteGenerator(number_of_colors=len(colors),
                                     rgb_colors=colors)

    def get_cluster_hulls(self):
        if not self.clusters.groups:
            return None
        var = self.clusters.table.domain["Clusters"]
        return [(hull, var.colors[var.values.index(key)])
                for key, (_, _, hull) in self.clusters.groups.items()]

    def get_cluster_labels(self):
        if not self.clusters.groups:
            return None
        var = self.clusters.table.domain["Clusters"]
        return [(ann, lab, var.colors[var.values.index(key)])
                for key, (ann, lab, _) in self.clusters.groups.items()]

    def _get_projection_data(self):
        if not self.scores.table or not self.embedding or not self.clusters.table:
            return self.data
        n_sco = self.scores.table.X.shape[1]
        n_emb = self.embedding.X.shape[1]
        n_clu = self.clusters.table.X.shape[1]
        domain = self.data.domain
        metas = chain(self.scores.table.domain.attributes,
                      self.embedding.domain.attributes,
                      self.clusters.table.domain.attributes,
                      domain.metas)
        domain = Domain(domain.attributes, domain.class_vars, metas)

        augmented_data = self.data.transform(domain)
        augmented_data.metas[:, :n_sco] = self.scores.table.X
        augmented_data.metas[:, n_sco:n_sco + n_emb] = self.embedding.X
        ind = n_sco + n_emb + n_clu
        augmented_data.metas[:, n_sco + n_emb: ind] = self.clusters.table.X
        augmented_data.metas[:, ind:] = self.data.metas
        return augmented_data

    def _get_send_report_caption(self):
        color_vr_name = "Clusters" if self.color_by_cluster else \
            self._get_caption_var_name(self.attr_color)
        return report.render_items_vert((
            ("Color", color_vr_name),
            ("Label", self._get_caption_var_name(self.attr_label)),
            ("Shape", self._get_caption_var_name(self.attr_shape)),
            ("Size", self._get_caption_var_name(self.attr_size)),
            ("Jittering", self.graph.jitter_size != 0 and
             "{} %".format(self.graph.jitter_size))))

    def clear(self):
        super().clear()
        self.cancel()
        self.embedding = None
        self.__invalidate_scores()
        self.__invalidate_clusters()

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()


if __name__ == "__main__":
    from orangecontrib.bioinformatics.utils import serverfiles

    data_path = "https://datasets.orange.biolab.si/sc/aml-1k.tab.gz"
    table_data = Table(data_path)
    table_data.attributes[TAX_ID] = "9606"
    genes_path = serverfiles.localpath_download(
        "marker_genes", "panglao_gene_markers.tab")
    f = FilterString("Organism", FilterString.Equal, "Human")
    table_genes = Values([f])(Table(genes_path))

    WidgetPreview(OWAnnotateProjection).run(
        set_data=table_data,
        set_subset_data=table_data[::10],
        set_genes=table_genes,
        # set_projector=PCA()
    )



"""
NCBI GEO module


GEO datasets are hosted on datasets.biolab.si and are not part of
a bioinformatics knowledge base (we can't use bioinformatics.utils.serverfiles module).

"""
import os
import json

from serverfiles import ServerFiles, LocalFiles
from Orange.misc.environ import data_dir
from Orange.data import Table, DiscreteVariable, Domain, filter as table_filter

from orangecontrib.bioinformatics.ncbi.gene import OrangeTableAnnotations as Annotations

domain = 'geo'
_local_cache_path = os.path.join(data_dir(), domain)
_all_info_file = os.path.join(_local_cache_path, '__INFO__')
_server_url = 'http://download.biolab.si/datasets/geo/'
pubmed_url = 'http://www.ncbi.nlm.nih.gov/pubmed/{}'

server_files = ServerFiles(server=_server_url)
local_files = LocalFiles(_local_cache_path, serverfiles=server_files)


def is_cached(gds_id):
    return os.path.exists(os.path.join(_local_cache_path, gds_id + '.tab'))


def info_cache(f):
    """ Store content of __INFO__ file locally.
    """

    def wrapper():
        if not os.path.isfile(_all_info_file) or os.path.getsize(_all_info_file) == 0:
            with open(_all_info_file, 'w') as fp:
                json.dump(list(server_files.allinfo().items()), fp)

        return f()

    return wrapper


@info_cache
def dataset_all_info():
    """
    Return a list of all GEO datasets from databases.biolab.si.
    """

    with open(_all_info_file, 'r') as fp:
        return json.load(fp)


def dataset_download(gds_id, samples=None, transpose=False, callback=None):
    file_name = '{}.tab'.format(gds_id)
    local_files.update(file_name, extract=True, callback=callback)

    table = Table(local_files.localpath_download(file_name))
    title = table.name
    gds_info = local_files.info(file_name)
    table_annotations = {Annotations.tax_id: gds_info['taxid']}

    if callback:
        callback()

    if samples is not None:
        filters = [table_filter.FilterStringList(sample, sample_types) for sample, sample_types in samples.items()]
        table = table_filter.Values(filters)(table)

        column_values = []
        for meta_var in samples.keys():
            column_values.append(table.get_column_view(table.domain[meta_var])[0])

        class_values = list(map('|'.join, zip(*column_values)))

        _class_values = list(set(class_values))
        map_class_values = dict((value, key) for (key, value) in enumerate(_class_values))
        class_var = DiscreteVariable(name='class', values=_class_values)
        _domain = Domain(table.domain.attributes,
                         table.domain.class_vars + (class_var, ),
                         table.domain.metas)

        table = table.transform(_domain)
        col, _ = table.get_column_view(class_var)
        col[:] = [map_class_values[class_val] for class_val in class_values]

    if transpose:
        table = Table.transpose(table, feature_names_column='sample_id', meta_attr_name='genes')
        table.name = title  # table name is lost after transpose
        table_annotations[Annotations.gene_as_attribute_name] = not gds_info[Annotations.gene_as_attribute_name]
        table_annotations[Annotations.gene_id_column] = gds_info[Annotations.gene_id_attribute]
    else:
        table_annotations[Annotations.gene_as_attribute_name] = gds_info[Annotations.gene_as_attribute_name]
        table_annotations[Annotations.gene_id_attribute] = gds_info[Annotations.gene_id_attribute]

    if callback:
        callback()

    table.attributes = table_annotations
    return table



""" GeneSets """
import threading
import concurrent.futures

from functools import partial

from AnyQt.QtWidgets import (
    QTreeView, QHeaderView, QHBoxLayout
)
from AnyQt.QtCore import (
    Qt, Slot, QThread
)
from AnyQt.QtGui import (
    QColor, QStandardItemModel, QStandardItem,
)

from Orange.widgets.gui import (
    vBox, lineEdit, LinkRole, LinkStyledItemDelegate, doubleSpin,
    auto_commit, widgetLabel, spin,  widgetBox, radioButtonsInBox
)
from Orange.data import Table
from Orange.widgets.settings import Setting
from Orange.widgets.utils.signals import Output, Input

from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID, GENE_AS_ATTRIBUTE_NAME, GENE_ID_COLUMN, GENE_ID_ATTRIBUTE
)

from orangecontrib.bioinformatics.widgets.utils.gui import GeneSetsSelection, NumericalColumnDelegate, FilterProxyModel
from orangecontrib.bioinformatics.widgets import OWGeneSets as gene_sets
from orangecontrib.bioinformatics.utils.statistics import FDR


class OWGeneSetEnrichment(gene_sets.OWGeneSets):
    name = "Gene Set Enrichment"
    description = ""
    icon = "icons/OWGeneSets.svg"
    priority = 9

    max_p_value = Setting(0.0001)
    use_p_value = Setting(False)
    max_fdr = Setting(0.01)
    use_max_fdr = Setting(True)
    use_reference_data = Setting(True, schema_only=True)

    COUNT, REFERENCE, P_VAL, FDR, ENRICHMENT, GENES, CATEGORY, TERM = range(8)
    DATA_HEADER_LABELS = ["Count", 'Reference', 'p-Value', 'FDR', 'Enrichment', 'Genes In Set', 'Category', 'Term']

    class Inputs(gene_sets.OWGeneSets.Inputs):
        reference = Input("Reference Genes", Table)

    class Outputs:
        matched_genes = Output("Matched Genes", Table)

    def __init__(self):
        # reference data attributes
        self.reference_data = None
        self.reference_genes = None
        self.reference_tax_id = None
        self.reference_attr_names = None
        self.reference_gene_id_attribute = None
        self.reference_gene_id_column = None
        self.reference_radio_box = None

        super().__init__()


    @Inputs.reference
    def handle_reference_genes(self, data):
        """
        Set the (optional) input dataset with reference gene names.
        """

        if data:
            self.reference_data = data
            self.reference_tax_id = str(self.reference_data.attributes.get(TAX_ID, None))
            self.reference_attr_names = self.reference_data.attributes.get(GENE_AS_ATTRIBUTE_NAME, None)
            self.reference_gene_id_attribute = self.reference_data.attributes.get(GENE_ID_ATTRIBUTE, None)
            self.reference_gene_id_column = self.reference_data.attributes.get(GENE_ID_COLUMN, None)

            if not (self.reference_attr_names is not None
                    and ((self.reference_gene_id_attribute is None) ^ (self.reference_gene_id_column is None))):

                if self.reference_tax_id is None:
                    self.Error.missing_annotation()
                    return

                self.Error.missing_gene_id()
                return

            elif self.reference_tax_id is None:
                self.Error.missing_tax_id()
                return

        self.__get_reference_genes()
        self.reference_radio_box.setEnabled(bool(self.reference_data))
        self.invalidate()

    def __get_source_data(self, proxy_row_index, column):
        proxy_index = self.filter_proxy_model.index(proxy_row_index, column)
        source_index = self.filter_proxy_model.mapToSource(proxy_index)
        return source_index.data(role=Qt.DisplayRole)

    def _update_fdr(self):
        # Update the FDR in place due to a changed selected categories set and
        # results for all of these categories are already available.
        proxy = self.filter_proxy_model
        model = self.filter_proxy_model.sourceModel()

        if model is not None:
            assert isinstance(model, QStandardItemModel)

            p_values = [(i, self.__get_source_data(i, self.P_VAL)) for i in range(proxy.rowCount())]
            fdr_values = FDR([p_val for _, p_val in p_values])

            for i, fdr_val in zip([i for i, _ in p_values], fdr_values):
                proxy_index = proxy.index(i, self.FDR)
                source_index = self.filter_proxy_model.mapToSource(proxy_index)
                source_item = model.item(source_index.row(), self.FDR)
                source_item.setData(fdr_val, role=Qt.DisplayRole)
                source_item.setData(fdr_val, role=Qt.ToolTipRole)

    def __get_reference_genes(self):
        self.reference_genes = []

        if self.reference_attr_names:
            for variable in self.reference_data.domain.attributes:
                self.reference_genes.append(str(variable.attributes.get(self.reference_gene_id_attribute, '?')))
        else:
            genes, _ = self.reference_data.get_column_view(self.reference_gene_id_column)
            self.reference_genes = [str(g) for g in genes]

    def create_filters(self):
        search_term = self.search_pattern.lower().strip().split()

        # apply filtering rules
        filters = [
            FilterProxyModel.Filter(
                self.TERM, Qt.DisplayRole,
                lambda value: all(fs in value.lower() for fs in search_term))
        ]

        if self.use_min_count:
           filters.append(
               FilterProxyModel.Filter(
                   self.COUNT, Qt.DisplayRole,
                   lambda value: value >= self.min_count,
               )
           )

        if self.use_p_value:
            filters.append(
                FilterProxyModel.Filter(
                    self.P_VAL, Qt.DisplayRole,
                    lambda value: value < self.max_p_value
                )
            )

        if self.use_max_fdr:
            filters.append(
                FilterProxyModel.Filter(
                    self.FDR, Qt.DisplayRole,
                    lambda value: value < self.max_fdr
                )
            )

        return filters

    def create_partial(self):
        reference_genes = self.reference_genes if (self.use_reference_data and self.reference_data)\
                                               else self.gs_widget.gs_object.genes()

        return partial(self.set_items,
                       self.gs_widget.gs_object,
                       self.stored_gene_sets_selection,
                       set(self.input_genes),
                       self.callback,
                       reference_genes=reference_genes)

    @staticmethod
    def set_items(gene_sets, sets_to_display, genes, callback, reference_genes=None):
        model_items = []
        if not genes:
            return

        for gene_set in sorted(gene_sets):
            if gene_set.hierarchy not in sets_to_display:
                continue

            reference_genes = [] if reference_genes is None else reference_genes
            enrichemnt_result = gene_set.set_enrichment(reference_genes, genes.intersection(reference_genes))
            callback()

            if len(enrichemnt_result.query) > 0:
                category_column = QStandardItem()
                name_column = QStandardItem()
                count_column = QStandardItem()
                genes_column = QStandardItem()
                ref_column = QStandardItem()
                pval_column = QStandardItem()
                fdr_column = QStandardItem()
                enrichment_column = QStandardItem()

                category_column.setData(", ".join(gene_set.hierarchy), Qt.DisplayRole)
                name_column.setData(gene_set.name, Qt.DisplayRole)
                name_column.setData(gene_set.name, Qt.ToolTipRole)
                name_column.setData(gene_set.link, LinkRole)
                name_column.setForeground(QColor(Qt.blue))

                count_column.setData(len(enrichemnt_result.query), Qt.DisplayRole)
                count_column.setData(set(enrichemnt_result.query), Qt.UserRole)

                genes_column.setData(len(gene_set.genes), Qt.DisplayRole)
                genes_column.setData(set(gene_set.genes), Qt.UserRole)  # store genes to get then on output on selection

                ref_column.setData(len(enrichemnt_result.reference), Qt.DisplayRole)

                pval_column.setData(enrichemnt_result.p_value, Qt.DisplayRole)
                pval_column.setData(enrichemnt_result.p_value, Qt.ToolTipRole)

                enrichment_column.setData(enrichemnt_result.enrichment_score, Qt.DisplayRole)
                enrichment_column.setData(enrichemnt_result.enrichment_score, Qt.ToolTipRole)

                model_items.append([count_column, ref_column, pval_column, fdr_column, enrichment_column,
                                    genes_column, category_column, name_column])
        return model_items

    # We must extend this, because we need to update FDR values after workers finish enrichment
    @Slot(concurrent.futures.Future)
    def _init_gene_sets_finished(self, f):
        assert self.thread() is QThread.currentThread()
        assert threading.current_thread() == threading.main_thread()
        assert self._task is not None
        assert self._task.future is f
        assert f.done()

        self._task = None
        self.progress_bar.finish()
        self.setStatusMessage('')

        try:
            results = f.result()  # type: list
            [self.data_model.appendRow(model_item) for model_item in results]
            self.filter_proxy_model.setSourceModel(self.data_model)
            self.data_view.selectionModel().selectionChanged.connect(self.commit)
            self._update_fdr()
            self.filter_data_view()
            self.set_selection()
            self.update_info_box()

        except Exception as ex:
            print(ex)

    def assign_delegates(self):
        self.data_view.setItemDelegateForColumn(
            self.GENES, NumericalColumnDelegate(self)
        )

        self.data_view.setItemDelegateForColumn(
            self.COUNT, NumericalColumnDelegate(self)
        )

        self.data_view.setItemDelegateForColumn(
            self.REFERENCE, NumericalColumnDelegate(self)
        )

        self.data_view.setItemDelegateForColumn(
            self.P_VAL, NumericalColumnDelegate(self, precision=2, notation='e')
        )

        self.data_view.setItemDelegateForColumn(
            self.FDR, NumericalColumnDelegate(self, precision=2, notation='e')
        )

        self.data_view.setItemDelegateForColumn(
            self.ENRICHMENT, NumericalColumnDelegate(self, precision=1)
        )

    def setup_control_area(self):
        # Control area
        self.input_info = widgetLabel(
            widgetBox(self.controlArea, "Info", addSpace=True), 'No data on input.\n'
        )
        self.custom_gs_col_box = box = vBox(self.controlArea, 'Custom Gene Set Term Column')
        box.hide()

        self.reference_radio_box = radioButtonsInBox(
            self.controlArea, self, "use_reference_data", ["Entire genome", "Reference gene set (input)"],
            tooltips=["Use entire genome (for gene set enrichment)", "Use reference set of genes"],
            box="Reference", callback=self.invalidate)

        self.reference_radio_box.setEnabled(False)

        gene_sets_box = widgetBox(self.controlArea, "Gene Sets")
        self.gs_widget = GeneSetsSelection(gene_sets_box, self, 'stored_gene_sets_selection')
        self.gs_widget.hierarchy_tree_widget.itemClicked.connect(self.update_tree_view)

        self.commit_button = auto_commit(self.controlArea, self, "auto_commit", "&Commit", box=False)

    def setup_filter_area(self):
        h_layout = QHBoxLayout()
        h_layout.setSpacing(100)
        h_widget = widgetBox(self.mainArea, orientation=h_layout)

        spin(h_widget, self, 'min_count', 0, 100,
             label='Count',
             tooltip='Minimum genes count',
             checked='use_min_count',
             callback=self.filter_data_view,
             callbackOnReturn=True,
             checkCallback=self.filter_data_view)

        doubleSpin(h_widget, self, 'max_p_value', 0.0, 1.0, 0.0001,
                   label='p-value',
                   tooltip='Maximum p-value of the enrichment score',
                   checked='use_p_value',
                   callback=self.filter_data_view,
                   callbackOnReturn=True,
                   checkCallback=self.filter_data_view
                   )

        doubleSpin(h_widget, self, 'max_fdr', 0.0, 1.0, 0.0001,
                   label='FDR',
                   tooltip='Maximum false discovery rate',
                   checked='use_max_fdr',
                   callback=self.filter_data_view,
                   callbackOnReturn=True,
                   checkCallback=self.filter_data_view
                   )

        self.line_edit_filter = lineEdit(h_widget, self, 'search_pattern')
        self.line_edit_filter.setPlaceholderText('Filter gene sets ...')
        self.line_edit_filter.textChanged.connect(self.filter_data_view)

    def setup_gui(self):
        # control area
        self.setup_control_area()

        # main area
        self.data_view = QTreeView()
        self.setup_filter_model()
        self.setup_filter_area()
        self.data_view.setAlternatingRowColors(True)
        self.data_view.sortByColumn(self.COUNT, Qt.DescendingOrder)
        self.data_view.setSortingEnabled(True)
        self.data_view.setSelectionMode(QTreeView.ExtendedSelection)
        self.data_view.setEditTriggers(QTreeView.NoEditTriggers)
        self.data_view.viewport().setMouseTracking(False)
        self.data_view.setItemDelegateForColumn(self.TERM, LinkStyledItemDelegate(self.data_view))

        self.mainArea.layout().addWidget(self.data_view)

        self.data_view.header().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.assign_delegates()


if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication
    app = QApplication([])
    ow = OWGeneSetEnrichment()
    ow.show()
    app.exec_()
